from __future__ import annotations

import numpy as np

from typing import List, Dict, Optional, Any, Union
from copy import deepcopy

from edgaro.data.dataset import Dataset
from edgaro.data.dataset_array import DatasetArray
from edgaro.base.base_transformer import BaseTransformer
from edgaro.base.utils import print_unbuffered


class BaseTransformerArray:
    """
    Create a class to apply BaseTransformer transformation with more than one set of parameters and/or
    to each of the Dataset objects in DatasetArray.

    Parameters
    ----------
    base_transformer : BaseTransformer
        The object defining the transformation procedure.
    parameters : list[list, Dict[str, Any]]], optional
        The list of parameters for base_transformer. If the object is used for a DatasetArray object,
        the parameter list should be nested. For details, see Examples section.
    transformer_sufix : str
        A suffix to be added to name in a transformed object.

    Attributes
    ----------
    transformer_sufix : str
        A suffix to be added to name in a transformed object.

    Examples
    ----------
    Example 1

    >>> from test.resources.objects import *
    >>> from edgaro.data.dataset import Dataset
    >>> from edgaro.data.dataset_array import DatasetArray
    >>> from edgaro.balancing.transformer import RandomUnderSampler
    >>> from edgaro.base.base_transformer_array import BaseTransformerArray
    >>> df = Dataset(name_1, df_1, target_1)
    >>> params = [{'sampling_strategy': 0.98}, {'sampling_strategy': 1}]
    >>> transformer = RandomUnderSampler()
    >>> array = BaseTransformerArray(transformer, parameters=params)
    >>> array.fit(df)
    >>> array.transform(df)

    Example 2

    >>> from test.resources.objects import *
    >>> from edgaro.data.dataset import Dataset
    >>> from edgaro.data.dataset_array import DatasetArray
    >>> from edgaro.balancing.transformer import RandomUnderSampler
    >>> from edgaro.base.base_transformer_array import BaseTransformerArray
    >>> df = DatasetArray([Dataset(name_2, df_1, target_1), Dataset(name_1, df_1, target_1)])
    >>> params = [ [{'sampling_strategy': 0.98}, {'sampling_strategy': 1}] for _ in range(len(df)) ]
    >>> transformer = RandomUnderSampler()
    >>> array = BaseTransformerArray(transformer, parameters=params)
    >>> array.fit(df)
    >>> array.transform(df)
    """

    def __init__(self, base_transformer: BaseTransformer, parameters: Optional[List[Union[List, Dict[str, Any]]]] = None,
                 transformer_sufix: str = '_transformed_array') -> None:
        self.__base_transformer = base_transformer
        self.__transformers = []
        self.__input_shape = None
        self.__parameters = parameters
        self.__was_fitted = False
        self.transformer_sufix = transformer_sufix

    def __create_new_transformer(self, param: Dict[str, Any]) -> BaseTransformer:
        t = deepcopy(self.__base_transformer)
        t.set_params(**param)
        return t

    def fit(self, dataset: Union[Dataset, DatasetArray]) -> None:
        """
        Fit the transformer.

        Parameters
        ----------
        dataset : Dataset, DatasetArray
            The object to fit BaseTransformerArray on.
        """
        # Single dataset case
        if isinstance(dataset, Dataset):
            self.__input_shape = 1
            if self.__parameters is None:
                self.__transformers = [deepcopy(self.__base_transformer)]
                self.__transformers[0].fit(dataset)
            else:
                self.__transformers = [self.__create_new_transformer(params) for params in self.__parameters]
                for i in range(len(self.__transformers)):
                    self.__transformers[i].fit(dataset)
        # DatasetArray case
        else:
            self.__input_shape = len(dataset)
            if self.__parameters is None:
                self.__transformers = [
                    deepcopy(self.__base_transformer)
                    if isinstance(dataset[i], Dataset)
                    else BaseTransformerArray(deepcopy(self.__base_transformer))
                    for i in range(len(dataset))
                ]
                for i in range(len(self.__transformers)):
                    self.__transformers[i].fit(dataset[i])
            else:
                if not len(dataset) == len(self.__parameters):
                    raise Exception('Not enough parameters were provided!')

                self.__transformers = [deepcopy(BaseTransformerArray(
                    base_transformer=self.__base_transformer,
                    parameters=BaseTransformerArray.__transform_dict_to_lst(self.__parameters[i]) if not isinstance(self.__parameters[i], list) else self.__parameters[i]))
                    for i in range(len(dataset))]

                for i in range(len(dataset)):
                    self.__transformers[i].fit(dataset[i])
        self.__was_fitted = True

    def transform(self, dataset: Union[Dataset, DatasetArray]) -> DatasetArray:
        """
        Transform the object.

        Parameters
        ----------
        dataset : Dataset, DatasetArray
            The object to be transformed.

        Returns
        -------
        DatasetArray
            The transformed object.
        """
        # Single dataset case
        if isinstance(dataset, Dataset):
            if not self.__input_shape == 1:
                raise Exception('DatasetArray was fitted, but single Dataset was provided!')

            tab = []
            for transformator in self.__transformers:
                try:
                    t = transformator.transform(dataset)
                    tab.append(t)
                except (Exception,) as exc:
                    print_unbuffered(f"An Exception occurred while transforming {dataset.name} with "
                                     f"{transformator.__str__()}")
                    print_unbuffered(exc)

            try:
                return DatasetArray(
                    tab,
                    name=dataset.name + self.transformer_sufix
                )
            except (Exception,):
                for i in range(len(tab)):
                    tab[i].name += ('_' + str(i))
                return DatasetArray(
                    tab,
                    name=dataset.name + self.transformer_sufix
                )
        # DatasetArray case
        else:
            if not self.__input_shape == len(dataset):
                raise Exception('Dataset was fitted, but DatasetArray was provided!')

            tab = []
            for i in range(len(dataset)):
                try:
                    t = self.__transformers[i].transform(dataset[i])
                    tab.append(t)
                except (Exception,) as exc:
                    print_unbuffered(f"An Exception occurred while transforming {dataset[i].name} with "
                                     f"{self.__transformers[i].__str__()}")
                    print_unbuffered(exc)

            return DatasetArray(
                tab,
                name=dataset.name + self.transformer_sufix
            )

    def set_params(self, **params) -> None:
        """
        Set params for BaseTransformerArray.

        The parameters should be in a form of a list or a nested list - it depends on whether the object is used
        on a Dataset or a DatasetArray object. For details see Examples section.

        Parameters
        ----------
        params : dict
            The parameters to be set.

        Examples
        ----------
        Example 1

        >>> from test.resources.objects import *
        >>> from edgaro.balancing.transformer import RandomUnderSampler
        >>> df = Dataset(name_1, df_1, target_1)
        >>> transformer = RandomUnderSampler()
        >>> array = BaseTransformerArray(transformer)
        >>> params = {'sampling_strategy': [1, 0.98]}
        >>> array.set_params(**params)
        >>> array.fit(df)
        >>> array.transform(df)

        Example 2

        >>> from test.resources.objects import *
        >>> from edgaro.balancing.transformer import RandomUnderSampler
        >>> df = DatasetArray([Dataset(name_2, df_1, target_1), Dataset(name_1, df_1, target_1)])
        >>> transformer = RandomUnderSampler()
        >>> array = BaseTransformerArray(transformer)
        >>> params = {'sampling_strategy': [[1, 0.98] for _ in range(len(df))]}
        >>> array.set_params(**params)
        >>> array.fit(df)
        >>> array.transform(df)
        """
        lengths = [len(val) for key, val in params.items()]
        if len(lengths) == 0:
            raise Exception('Parameters were not provided!')

        if not np.alltrue(np.array(lengths) == lengths[0]):
            raise Exception('Parameters do not have the same length!')

        self.__parameters = BaseTransformerArray.__transform_dict_to_lst(params)

        if len(self.__transformers) != 0:
            for i in range(len(self.__transformers)):
                self.__transformers[i].set_params(**self.__parameters[i])

    @staticmethod
    def __transform_dict_to_lst(dct):
        if np.alltrue([not isinstance(val, list) for key, val in dct.items()]):
            return [dct]
        lengths = [len(val) for key, val in dct.items()]
        tmp = []
        for i in range(lengths[0]):
            tmp_dict = {}
            for key in dct:
                tmp_dict[key] = dct[key][i]
            tmp.append(tmp_dict)
        return tmp

    def get_params(self) -> Optional[List[Dict[str, Any]]]:
        """
        Get parameters of BaseTransformerArray.

        Returns
        -------
        list[list, Dict[str, Any]]], optional
            The parameters.
        """
        return self.__parameters

    @property
    def was_fitted(self) -> bool:
        """
        The information whether the BaseTransformerArray was fitted.

        Returns
        -------
        bool
        """
        if self.__was_fitted:
            return True
        elif len(self.__transformers) == 0:
            return False
        else:
            return np.alltrue([
                transformer.was_fitted for transformer in self.__transformers
            ])

    @property
    def parameters(self) -> Optional[List[Dict[str, Any]]]:
        """
        Get parameters of BaseTransformerArray.

        Returns
        -------
        list[list, Dict[str, Any]]], optional
            The parameters.
        """
        return self.__parameters

    @parameters.setter
    def parameters(self, val: Optional[List[Dict[str, Any]]]) -> None:
        if not self.was_fitted:
            self.__parameters = val
        else:
            raise Exception('Parameters were not set since Transformer has already been fitted!')

    @property
    def transformers(self) -> List[Union[BaseTransformer, BaseTransformerArray, List]]:
        """
        BaseTransformer and BaseTransformerArray objects inside this object.

        Returns
        -------
        list[BaseTransformer, BaseTransformerArray, list]
        """
        return self.__transformers

    @transformers.setter
    def transformers(self, val: List[Union[BaseTransformer, BaseTransformerArray, List]]) -> None:
        if not self.was_fitted:
            self.__transformers = val
        else:
            raise Exception('Transformers were not set since Transformer has already been fitted!')

    @property
    def base_transformer(self) -> BaseTransformer:
        """
        BaseTransformer of this object.

        Returns
        -------
        BaseTransformer
        """
        return self.__base_transformer

    @base_transformer.setter
    def base_transformer(self, val: BaseTransformer):
        if not self.was_fitted:
            self.__transformers = val
        else:
            raise Exception('Base transformers were not set since Transformer has already been fitted!')

    def __len__(self) -> int:
        return len(self.__transformers)

    def __getitem__(self, key: Union[int, List[int]]) -> Optional[
                    Union[BaseTransformer, BaseTransformerArray, List]]:
        if isinstance(key, list):
            out = [self.__getitem__(k) for k in key]
            out = [o for o in out if o is not None]
            if len(out) == 0:
                return None
            else:
                return out
        elif isinstance(key, int):
            if key <= len(self.__transformers):
                return self.__transformers[key]
        return None

    def __str__(self) -> str:
        return f"BaseTransformerArray{(' ' + self.__class__.__name__) if self.__class__.__name__ != 'BaseTransformerArray' else ''} with {len(self.transformers)} transformers"

    def __repr__(self) -> str:
        return f"<BaseTransformerArray{(' ' + self.__class__.__name__) if self.__class__.__name__ != 'BaseTransformerArray' else ''} with {len(self.transformers)} transformers>"



import random
import string
import pandas as pd
import pytest

from copy import deepcopy
from sklearn.ensemble import RandomForestClassifier

from edgaro.data.dataset import Dataset, DatasetFromOpenML
from edgaro.model.model import RandomForest, ModelFromSKLEARN, XGBoost, RandomSearchCV, GridSearchCV

from .resources.objects import *


@pytest.fixture(scope='module', params=[
    DatasetFromOpenML(task_id=task_id_2, apikey=APIKEY),
    Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)]))
])
def ds_test_fitting(request):
    ds = request.param
    ds = deepcopy(ds)
    ds.remove_nans()
    return ds


@pytest.fixture(scope='module', params=[
    DatasetFromOpenML(task_id=task_id_2, apikey=APIKEY),
    Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)]))
])
def ds_test_fitting_with_split(request):
    ds = request.param
    ds = deepcopy(ds)
    ds.remove_nans()
    ds.train_test_split(test_size=0.3, random_state=42)
    return ds


class TestFitting:
    def test_random_forest(self, ds_test_fitting):
        try:
            model = RandomForest(test_size=None, max_depth=1, n_estimators=1, random_state=42)
            model.fit(ds_test_fitting)
            model.evaluate()
            model.predict(ds_test_fitting)
            model.predict_proba(ds_test_fitting)
            str(model)
            repr(model)
        except (Exception,):
            assert False

        with pytest.raises(Exception):
            model.predict(ds_test_fitting.test)
            model.predict_proba(ds_test_fitting.test)

    def test_fit_verbose(self, ds_test_fitting, capsys):
        model = RandomForest(test_size=None, max_depth=1, n_estimators=1, random_state=42, verbose=True)
        model.fit(ds_test_fitting)
        captured = capsys.readouterr()
        assert f'Model {RandomForest().__repr__()} is being fitted with {ds_test_fitting.name}' in captured.out
        assert f'Model {model.__repr__()} was fitted with {ds_test_fitting.name}' in captured.out

    def test_fit_print_scores(self, ds_test_fitting, capsys):
        model = RandomForest(test_size=None, max_depth=1, n_estimators=1, random_state=42)
        model.fit(ds_test_fitting, print_scores=True)
        captured = capsys.readouterr()
        assert not captured.out == ''

    def test_evaluate_verbose(self, ds_test_fitting, capsys):
        model = RandomForest(test_size=None, max_depth=1, n_estimators=1, random_state=42)
        model.fit(ds_test_fitting)
        model.verbose = True
        model.evaluate()
        captured = capsys.readouterr()
        assert f'Model {model.__repr__()} is being evaluated' in captured.out
        assert f'Model {model.__repr__()} was evaluated' in captured.out

    def test_xgboost(self, ds_test_fitting):
        try:
            model = XGBoost(test_size=None, max_depth=1, n_estimators=1, random_state=42)
            model.fit(ds_test_fitting)
            model.evaluate()
            model.predict(ds_test_fitting)
            model.predict_proba(ds_test_fitting)
            str(model)
            repr(model)
        except (Exception,):
            assert False

        with pytest.raises(Exception):
            model.predict(ds_test_fitting.test)
            model.predict_proba(ds_test_fitting.test)

    def test_grid_search(self, ds_test_fitting):
        try:
            model = GridSearchCV(test_size=None, base_model=RandomForest(),
                                 param_grid={'n_estimators': [1, 2], 'max_depth': [1]}, random_state=42)
            model.fit(ds_test_fitting)
            model.evaluate()
            model.predict(ds_test_fitting)
            model.predict_proba(ds_test_fitting)
            str(model)
            repr(model)
        except (Exception,):
            assert False

        with pytest.raises(Exception):
            model.predict(ds_test_fitting.test)
            model.predict_proba(ds_test_fitting.test)

    def test_random_search(self, ds_test_fitting):
        try:
            model = RandomSearchCV(test_size=None, base_model=RandomForest(),
                                   param_grid={'n_estimators': [1, 2], 'max_depth': [1]}, random_state=42)
            model.fit(ds_test_fitting)
            model.evaluate()
            model.predict(ds_test_fitting)
            model.predict_proba(ds_test_fitting)
            str(model)
            repr(model)
        except (Exception,):
            assert False

        with pytest.raises(Exception):
            model.predict(ds_test_fitting.test)
            model.predict_proba(ds_test_fitting.test)

    def test_was_fitted(self, ds_test_fitting):
        model = RandomForest(test_size=None, max_depth=1, n_estimators=1, random_state=42)
        assert not model.was_fitted
        model.fit(ds_test_fitting)
        assert model.was_fitted


def test_random_forest_empty_data():
    ds = Dataset(name_1, None, target_1)
    model = RandomForest(test_size=None, max_depth=1, n_estimators=1, random_state=42)
    with pytest.raises(Exception):
        model.fit(ds)


def test_random_forest_empty_data2():
    ds = Dataset(name_1, df_1, target_1)
    ds.data = pd.DataFrame()
    model = RandomForest(test_size=None, max_depth=1, n_estimators=1, random_state=42)
    with pytest.raises(Exception):
        model.fit(ds)


def test_random_forest_empty_target():
    ds = Dataset(name_1, df_1, None)
    model = RandomForest(test_size=None, max_depth=1, n_estimators=1, random_state=42)
    with pytest.raises(Exception):
        model.fit(ds)


def test_random_forest_non_binary():
    ds = Dataset(name_1, df_1, target_1_fake)
    model = RandomForest(test_size=None, max_depth=1, n_estimators=1, random_state=42)
    with pytest.raises(Exception):
        model.fit(ds)


def test_random_forest_empty_target2():
    ds = Dataset(name_1, df_1, target_1)
    ds.target = pd.Series()
    model = RandomForest(test_size=None, max_depth=1, n_estimators=1, random_state=42)
    with pytest.raises(Exception):
        model.fit(ds)


class TestFittingWithSplit:
    def test_random_forest_with_test(self, ds_test_fitting):
        try:
            ds = deepcopy(ds_test_fitting)
            model = RandomForest(test_size=0.3, max_depth=1, n_estimators=1, random_state=42)
            model.fit(ds)
            model.evaluate()
            model.predict(ds)
            model.predict_proba(ds)
            str(model)
            repr(model)
            model.predict(ds.test)
            model.predict_proba(ds.test)
        except (Exception,):
            assert False

    def test_xgboost_with_test(self, ds_test_fitting):
        try:
            ds = deepcopy(ds_test_fitting)
            model = XGBoost(test_size=0.3, max_depth=1, n_estimators=1, random_state=42)
            model.fit(ds)
            model.evaluate()
            model.predict(ds)
            model.predict_proba(ds)
            str(model)
            repr(model)
            model.predict(ds.test)
            model.predict_proba(ds.test)
        except (Exception,):
            assert False

    def test_grid_search_with_test(self, ds_test_fitting):
        try:
            ds = deepcopy(ds_test_fitting)
            model = GridSearchCV(test_size=0.3, base_model=RandomForest(),
                                 param_grid={'n_estimators': [1, 2], 'max_depth': [1]}, random_state=42)
            model.fit(ds)
            model.evaluate()
            model.predict(ds)
            model.predict_proba(ds)
            str(model)
            repr(model)
            model.predict(ds.test)
            model.predict_proba(ds.test)
        except (Exception,):
            assert False

    def test_random_search_with_test(self, ds_test_fitting):
        try:
            ds = deepcopy(ds_test_fitting)
            model = RandomSearchCV(test_size=0.3, base_model=RandomForest(),
                                   param_grid={'n_estimators': [1, 2], 'max_depth': [1]}, random_state=42)
            model.fit(ds)
            model.evaluate()
            model.predict(ds)
            model.predict_proba(ds)
            str(model)
            repr(model)
            model.predict(ds.test)
            model.predict_proba(ds.test)
        except (Exception,):
            assert False


class TestFittingWithEarlierSplit:
    def test_random_forest(self, ds_test_fitting_with_split):
        try:
            model = RandomForest(test_size=None, max_depth=1, n_estimators=1, random_state=42)
            model.fit(ds_test_fitting_with_split)
            model.evaluate()
            model.predict(ds_test_fitting_with_split)
            model.predict_proba(ds_test_fitting_with_split)
            str(model)
            repr(model)
            model.predict(ds_test_fitting_with_split.test)
            model.predict_proba(ds_test_fitting_with_split.test)
        except (Exception,):
            assert False

    def test_xgboost(self, ds_test_fitting_with_split):
        try:
            model = XGBoost(test_size=None, max_depth=1, n_estimators=1, random_state=42)
            model.fit(ds_test_fitting_with_split)
            model.evaluate()
            model.predict(ds_test_fitting_with_split)
            model.predict_proba(ds_test_fitting_with_split)
            str(model)
            repr(model)
            model.predict(ds_test_fitting_with_split.test)
            model.predict_proba(ds_test_fitting_with_split.test)
        except (Exception,):
            assert False

    def test_grid_search(self, ds_test_fitting_with_split):
        try:
            model = GridSearchCV(test_size=None, base_model=RandomForest(),
                                 param_grid={'n_estimators': [1, 2], 'max_depth': [1]}, random_state=42)
            model.fit(ds_test_fitting_with_split)
            model.evaluate()
            model.predict(ds_test_fitting_with_split)
            model.predict_proba(ds_test_fitting_with_split)
            str(model)
            repr(model)
            model.predict(ds_test_fitting_with_split.test)
            model.predict_proba(ds_test_fitting_with_split.test)
        except (Exception,):
            assert False

    def test_random_search(self, ds_test_fitting_with_split):
        try:
            model = RandomSearchCV(test_size=None, base_model=RandomForest(),
                                   param_grid={'n_estimators': [1, 2], 'max_depth': [1]}, random_state=42)
            model.fit(ds_test_fitting_with_split)
            model.evaluate()
            model.predict(ds_test_fitting_with_split)
            model.predict_proba(ds_test_fitting_with_split)
            str(model)
            repr(model)
            model.predict(ds_test_fitting_with_split.test)
            model.predict_proba(ds_test_fitting_with_split.test)
        except (Exception,):
            assert False


def test_fitting_from_sklearn_model(ds_test_fitting_with_split):
    try:
        model = ModelFromSKLEARN(RandomForestClassifier(max_depth=1, n_estimators=1, random_state=42), test_size=0.3)
        model.fit(ds_test_fitting_with_split)
        model.predict(ds_test_fitting_with_split)
        model.predict_proba(ds_test_fitting_with_split)
        model.predict(ds_test_fitting_with_split.test)
        model.predict_proba(ds_test_fitting_with_split.test)
    except (Exception,):
        assert False


@pytest.mark.parametrize('model', [
    RandomForest(max_depth=1, n_estimators=1, random_state=42),
    XGBoost(max_depth=1, n_estimators=1, random_state=42)
])
class TestModel:
    def test_transform_data(self, model, ds_test_fitting):
        model = deepcopy(model)
        ds = deepcopy(ds_test_fitting)

        model.fit(ds)
        try:
            X = model.transform_data(ds)
        except (Exception,):
            assert False

    def test_transform_data_none(self, model, ds_test_fitting):
        model = deepcopy(model)
        ds = deepcopy(ds_test_fitting)

        model.fit(ds)
        ds.data = None
        with pytest.raises(Exception,):
            X = model.transform_data(ds)

    def test_transform_data_empty(self, model, ds_test_fitting):
        model = deepcopy(model)
        ds = deepcopy(ds_test_fitting)

        model.fit(ds)
        ds.data = pd.DataFrame()
        with pytest.raises(Exception,):
            X = model.transform_data(ds)

    def test_transform_target(self, model, ds_test_fitting):
        model = deepcopy(model)
        ds = deepcopy(ds_test_fitting)

        model.fit(ds)
        try:
            Y = model.transform_target(ds)
        except (Exception,):
            assert False

    def test_transform_target_none(self, model, ds_test_fitting):
        model = deepcopy(model)
        ds = deepcopy(ds_test_fitting)

        model.fit(ds)
        ds.target = None
        with pytest.raises(Exception,):
            X = model.transform_target(ds)

    def test_transform_target_empty(self, model, ds_test_fitting):
        model = deepcopy(model)
        ds = deepcopy(ds_test_fitting)

        model.fit(ds)
        ds.target = pd.Series()
        with pytest.raises(Exception,):
            X = model.transform_target(ds)

    def test_output_types(self, model, ds_test_fitting):
        model = deepcopy(model)
        ds = deepcopy(ds_test_fitting)
        model.fit(ds)

        y = model.predict(ds)
        assert isinstance(y, Dataset)
        assert y.check_binary_classification()
        y = model.predict_proba(ds)
        assert isinstance(y, Dataset)

    def test_output_predict_with_none(self, model, ds_test_fitting):
        model = deepcopy(model)
        ds = deepcopy(ds_test_fitting)
        model.fit(ds)

        ds.data = None

        with pytest.raises(Exception):
            y = model.predict(ds)

        with pytest.raises(Exception):
            y = model.predict_proba(ds)

    def test_output_verbose(self, model, ds_test_fitting, capsys):
        model = deepcopy(model)
        model.verbose = True
        ds = deepcopy(ds_test_fitting)
        model.fit(ds)

        y = model.predict(ds)
        y = model.predict_proba(ds)

        captured = capsys.readouterr()
        assert f'Model {model.__repr__()} predicted on {ds.name}' in captured.out
        assert f'Model {model.__repr__()} predicted probabilities on {ds.name}' in captured.out

    def test_output_names(self, model, ds_test_fitting):
        model = deepcopy(model)
        ds = deepcopy(ds_test_fitting)
        model.fit(ds)

        y = model.predict(ds)
        assert y.name == ds.name + '_predicted'
        y = model.predict_proba(ds)
        assert y.name == ds.name + '_predicted_probabilities'

    def test_output_names_with_model_name(self, model, ds_test_fitting):
        model_name = ''.join(random.choice(string.ascii_uppercase) for _ in range(5))
        model = deepcopy(model)
        model.name = model_name
        ds = deepcopy(ds_test_fitting)
        model.fit(ds)

        y = model.predict(ds)
        assert y.name == ds.name + '_' + model_name + '_predicted'
        y = model.predict_proba(ds)
        assert y.name == ds.name + '_' + model_name + '_predicted_probabilities'



from __future__ import annotations

import math
import re

from abc import ABC, abstractmethod
import matplotlib.pyplot as plt
import numpy as np

from typing import List, Literal, Optional, Union, Tuple
from matplotlib.axes import Axes
from statsmodels.stats.multitest import fdrcorrection

from edgaro.explain.explainer_result import ModelProfileExplanation, ModelPartsExplanation


class ExplanationArray(ABC):
    def __int__(self) -> None:
        pass

    @abstractmethod
    def plot(self) -> None:
        pass

    @abstractmethod
    def compare(self) -> List[Union[float, list]]:
        pass

    @abstractmethod
    def plot_summary(self) -> None:
        pass


class ModelProfileExplanationArray(ExplanationArray):
    """
    The class which represent the PDP/ALE curves for all variables in Model/ModelArray object.

    Parameters
    ----------
    results : list[ModelProfileExplanation, ModelProfileExplanationArray]
        A list of ModelProfileExplanation/ModelProfileExplanationArray with results.
    name : str
        The name of ModelProfileExplanationArray. It is best if it is a Model/ModelArray name.
    explanation_type : {'PDP', 'ALE'}, default='PDP'
        A curve type.

    Attributes
    ----------
    results : list[ModelProfileExplanation, ModelProfileExplanationArray]
        A list of ModelProfileExplanation/ModelProfileExplanationArray with results.
    name : str
        The name of ModelProfileExplanationArray. It is best if it is a Model/ModelArray name.
    explanation_type : {'PDP', 'ALE'}
        A curve type.

    """

    def __init__(self, results: List[Union[ModelProfileExplanation, ModelProfileExplanationArray]], name: str,
                 explanation_type: Literal['PDP', 'ALE'] = 'PDP') -> None:
        self.results = results
        self.name = name
        self.explanation_type = explanation_type

    def __len__(self) -> int:
        return len(self.results)

    def __getitem__(self, key: Union[Union[str, int], List[Union[str, int]]]) \
            -> Optional[Union[ModelProfileExplanation, ModelProfileExplanationArray]]:
        if isinstance(key, list):
            outs = [self.__getitem__(k) for k in key]
            outs = [o for o in outs if o is not None]
            if len(outs) == 0:
                return None
            else:
                return ModelProfileExplanationArray(results=outs, name=self.name + "_subset",
                                                    explanation_type=self.explanation_type)
        elif isinstance(key, str):
            for result in self.results:
                if result.name == key:
                    return result
        elif isinstance(key, int):
            if key <= len(self.results):
                return self.results[key]
        return None

    @staticmethod
    def __check_colnames(res_array, col_names_in):
        if isinstance(res_array, ModelProfileExplanation):
            if len(col_names_in) == 0:
                col_names_in += res_array.results.keys()
                return True
            else:
                return set(res_array.results.keys()) == set(col_names_in)
        else:
            for res in res_array.results:
                correct = ModelProfileExplanationArray.__check_colnames(res, col_names_in)
                if not correct:
                    return False
            return True

    @staticmethod
    def __find_matching_explainer_result_array(res_array, results_in, variables, model_filter):
        if isinstance(res_array, ModelProfileExplanation):
            if np.all([var in res_array.results.keys() for var in variables]):
                if model_filter is not None:
                    if re.search(model_filter, res_array.name) is not None:
                        results_in.append(res_array)
                else:
                    results_in.append(res_array)
        else:
            for res in res_array.results:
                ModelProfileExplanationArray.__find_matching_explainer_result_array(res, results_in, variables,
                                                                                    model_filter)

    def plot(self, variables: Optional[List[str]] = None, n_col: int = 3, figsize: Optional[Tuple[int, int]] = None,
             model_filter: Optional[str] = None, index_base: Union[str, int] = -1, centered: bool = False):
        """
        The function plots the PDP/ALE curves for given variables using all available Curves in the object.

        Parameters
        ----------
        index_base : int, str, default=-1
            Index of a curve to be a base for comparisons.
        variables : list[str], optional, default=None
            Variables for which the plot should be generated. If None, plots for all variables are generated if all the
            available ModelProfileExplanation objects have exactly the same set of column names.
        n_col : int, default=3
            Number of columns in the final plot.
        figsize : tuple(int, int), optional, default=None
            The size of a figure. If None, the figure size is calculates as (8 * n_col, 8 * n_rows).
        model_filter : str, optional, default=None
            A regex expression to filter the names of the ModelProfileExplanation objects for comparing.
        centered : bool, default = False
            If True, the plots will be centered to start at 0.

        """
        if variables is None:
            col_names = []
            if not ModelProfileExplanationArray.__check_colnames(self, col_names):
                raise Exception('The ResultArray contains curves for models with different column names!')
            variables = col_names

        results = []
        ModelProfileExplanationArray.__find_matching_explainer_result_array(self, results, variables, model_filter)

        if len(results) == 0:
            raise Exception('There are not matching models!')

        base_model = self.results[index_base]

        try:
            results.remove(base_model)
        except (Exception,):
            pass

        results.insert(0, base_model)

        n_rows = math.ceil(len(variables) / n_col)
        if figsize is None:
            figsize = (8 * n_col, 8 * n_rows)
        fig, axes = plt.subplots(n_rows, n_col, figsize=figsize)

        for i in range(len(variables)):
            if n_rows > 1:
                ax_to_pass = axes[math.floor(i / n_col)][i - n_col * math.floor(i / n_col)]
            else:
                if n_col > 1:
                    ax_to_pass = axes[math.floor(i / n_col)]
                else:
                    ax_to_pass = axes

            results[0].plot(
                add_plot=[results[j] for j in range(1, len(results))],
                variable=variables[i],
                ax=ax_to_pass,
                show_legend=False,
                centered=centered
            )

        fig.tight_layout(rect=[0, 0.05, 1, 0.97])
        fig.legend([x.name for x in results], ncol=n_col, loc='lower center')
        plt.suptitle(f"{self.explanation_type} curves for {self.name}", fontsize=18)

    def compare(self, variable: Optional[Union[str, List[str]]] = None, index_base: Union[str, int] = -1,
                return_raw: bool = True, return_raw_per_variable: bool = True, model_filter: Optional[str] = None) \
            -> List[Union[float, List]]:
        """
        The function compares the curves in the array.

        Parameters
        ----------
        variable : list[str], optional, default=None
            List of variable names to calculate the metric distances. If None, the metrics are calculated for
            all the columns in this object.
        index_base : int, str, default=-1
            Index of a curve to be a base for comparisons.
        return_raw : bool, default=True
            If True, the metrics for each of the model are returned. Otherwise, the mean of the values is returned.
        return_raw_per_variable : bool, default=True
            If True, the metrics for each of the variables are returned. Otherwise, the mean of the values is returned.
        model_filter : str, optional, default=None
            A regex expression to filter the names of the ModelProfileExplanation objects for comparing.

        Returns
        -------
        list[float, list]

        """

        if isinstance(self.results[index_base], ModelProfileExplanation):
            if isinstance(index_base, int) and index_base < 0:
                index_base = self.results.index(self.results[index_base])

            def filter_objects(obj):
                if model_filter is not None and \
                        re.search(model_filter, obj.name) is None:
                    return False
                return True

            def flatten(lst):
                out = []
                for i in range(len(lst)):
                    if not (isinstance(lst[i], list) or isinstance(lst[i], ModelProfileExplanationArray)):
                        out.append(lst[i])
                    else:
                        tmp = flatten(lst[i])
                        out = out + tmp
                return out

            base_model = self[index_base]
            if base_model is None:
                raise Exception('Wrong index_base argument!')

            res = flatten(self.results)
            res.remove(self.results[index_base])

            if return_raw:
                out = []
                for i in range(len(res)):
                    if not filter_objects(res[i]):
                        continue

                    if not return_raw_per_variable:
                        out.append(base_model.compare(res[i], variable=variable,
                                                      return_raw_per_variable=False)[0])
                    else:
                        out.append(base_model.compare(res[i], variable=variable,
                                                      return_raw_per_variable=True))
                return out
            else:
                tab = [res[i] for i in range(len(res)) if filter_objects(res[i])]
                return base_model.compare(tab, variable=variable, return_raw_per_variable=return_raw_per_variable)
        elif np.alltrue([isinstance(res, ModelProfileExplanationArray) for res in self.results]):
            return [
                res.compare(variable=variable, index_base=index_base, return_raw=return_raw,
                            return_raw_per_variable=return_raw_per_variable, model_filter=model_filter)
                for res in self.results
            ]
        else:
            raise Exception('Wrong result structure!')

    def plot_summary(self, model_filters: Optional[List[str]] = None, filter_labels: [List[str]] = None,
                     variables: Optional[List[str]] = None, figsize: Optional[Tuple[int, int]] = None,
                     index_base: Union[str, int] = -1, return_df: bool = False):
        """
        The function plots boxplots of comparison metrics of curves in the object.

        Parameters
        ----------
        variables : list[str], optional, default=None
            Variables for which the plot should be generated. If None, plots for all variables are generated if all the
            available ModelProfileExplanation objects have exactly the same set of column names.
        figsize : tuple(int, int), optional, default=None
            The size of a figure.
        model_filters : list[str], optional, default=None
            List of regex expressions to filter the names of the ModelProfileExplanation objects for comparing.
            Each element in the list creates a new boxplot. If None, one boxplot of all results is plotted.
        filter_labels : list[str], optional, default=None
            Labels of model filters.
        index_base : int, str, default=-1
            Index of a curve to be a base for comparisons.
        return_df : bool, default=False
            If True, the method returns a dataframe on which a plot is created.
        """

        fig, ax = plt.subplots(figsize=figsize)
        plt.title(f'Summary of {self.explanation_type} for {self.name}')
        plt.xlabel('Filter')
        plt.ylabel(r'SDD values [$10^{-3}$]')

        def format_func(value, tick_number):
            return str(int(value * 10 ** 3))

        def flatten(lst):
            out = []
            for i in range(len(lst)):
                if not isinstance(lst[i], list):
                    out.append(lst[i])
                else:
                    tmp = flatten(lst[i])
                    out = out + tmp
            return out

        ax.yaxis.set_major_formatter(plt.FuncFormatter(format_func))

        if model_filters is None:
            results = self.compare(variable=variables, index_base=index_base,
                                   return_raw=True, return_raw_per_variable=True)

            results = flatten(results)

            if filter_labels is not None and len(filter_labels) == 1 and \
                    filter_labels is not None and len(filter_labels) == 1:
                plt.boxplot(results, labels=filter_labels, patch_artist=True)
            else:
                plt.boxplot(results, labels=['All values'], patch_artist=True)

        else:
            results = []
            for f in model_filters:
                tmp_out = self.compare(variable=variables, index_base=index_base,
                                       return_raw=True, return_raw_per_variable=True, model_filter=f)
                tmp_out = flatten(tmp_out)
                results.append(tmp_out)

            if filter_labels is not None:
                if len(filter_labels) == len(model_filters):
                    plt.boxplot(results, labels=filter_labels, patch_artist=True)
                else:
                    raise Exception('Incorrect length of filter_labels!')
            else:
                plt.boxplot(results, patch_artist=True, labels=model_filters)

        if return_df:
            return results

    def __str__(self) -> str:
        return f"ModelProfileExplanationArray {self.name} for {len(self.results)} variables: {list(self.results)} with {self.explanation_type} curve type"

    def __repr__(self) -> str:
        return f"<ModelProfileExplanationArray {self.name} with {self.explanation_type} curve type>"


class ModelPartsExplanationArray(ExplanationArray):
    """
    The class which represent the Variable Importance for all variables in Model/ModelArray object.

    Parameters
    ----------
    results : list[ModelPartsExplanation, ModelPartsExplanationArray]
        A list of ModelPartsExplanation/ModelPartsExplanationArray with results.
    name : str
        The name of ModelPartsExplanationArray. It is best if it is a Model/ModelArray name.
    explanation_type : {'VI'}, default='VI'
        An explanation type.

    Attributes
    ----------
    results : list[ModelPartsExplanation, ModelPartsExplanationArray]
        A list of ModelPartsExplanation/ModelPartsExplanationArray with results.
    name : str
        The name of ModelPartsExplanationArray. It is best if it is a Model/ModelArray name.
    explanation_type : {'VI'}, default='VI'
        An explanation type.

    """

    def __init__(self, results: List[Union[ModelPartsExplanation, ModelPartsExplanationArray]],
                 name: str, explanation_type: Literal['VI'] = 'VI') -> None:
        self.results = results
        self.name = name
        self.explanation_type = explanation_type

    def __len__(self) -> int:
        return len(self.results)

    def __getitem__(self, key: Union[Union[str, int], List[Union[str, int]]]) \
            -> Optional[Union[ModelPartsExplanation, ModelPartsExplanationArray]]:
        if isinstance(key, list):
            outs = [self.__getitem__(k) for k in key]
            outs = [o for o in outs if o is not None]
            if len(outs) == 0:
                return None
            else:
                return ModelPartsExplanationArray(results=outs, name=self.name + "_subset",
                                                  explanation_type=self.explanation_type)
        elif isinstance(key, str):
            for result in self.results:
                if result.name == key:
                    return result
        elif isinstance(key, int):
            if key <= len(self.results):
                return self.results[key]
        return None

    def plot(self, variable: Optional[Union[str, List[str]]] = None, max_variables: Optional[int] = None,
             index_base: Union[str, int] = -1, figsize: Optional[Tuple[int, int]] = (8, 8),
             ax: Optional[Axes] = None, show_legend: bool = True, x_lim: Optional[Tuple[float, float]] = None,
             metric_precision: int = 3) -> None:
        """
        The function plots the Variable Importance profile using all ModelPartsExplanation objects.

        Parameters
        ----------
        variable : str, list[str], optional, default=None
            Variable for which the VI should be plotted. If None, the all columns is plotted.
        figsize : tuple(int, int), optional, default=(8, 8)
            Size of a figure.
        max_variables : int, optional, default=None
            Maximal number of variables from the current object to be taken into account.
        ax : matplotlib.axes.Axes, optional, default=None
            The parameter should be passed if the plot is to be created in a certain Axis. In that situation, `figsize`
            parameter is ignored.
        show_legend : bool, default=True
            The parameter indicates whether the legend should be plotted.
        x_lim : tuple(float, float), optional, default=None
            The limits of 0X axis.
        metric_precision : int, default=5
            Number of digits to round the value of the metric value.
        index_base : int, str, default=-1
            Index of an explanation to be a base for comparisons.
        """

        base = self.results[index_base]
        plots = self.results.copy()
        plots.remove(base)
        plots = ModelPartsExplanationArray.__flatten(plots)

        base.plot(variable=variable, figsize=figsize, max_variables=max_variables,
                  add_plot=plots, ax=ax, show_legend=show_legend, x_lim=x_lim, metric_precision=metric_precision)

    def compare(self, variable: Optional[Union[str, List[str]]] = None, max_variables: Optional[int] = None,
                return_raw: bool = True, index_base: Union[str, int] = -1, model_filter: Optional[str] = None) \
            -> List[Union[float, list]]:
        """
        The function compares variable importance in the array.

        Parameters
        ----------
        variable : str, list[str], optional, default=None
            List of variable names to calculate the metric distances. If None, the metrics are calculated for
            all the columns in this object.
        max_variables : int, optional, default=None
            Maximal number of variables from the current object to be taken into account.
        return_raw : bool, default=True
            If True, the p-values are returned for each model. Otherwise, the mean value is returned.
        index_base : int, str, default=-1
            Index of an explanation to be a base for comparisons.
        model_filter : str, optional, default=None
            A regex expression to filter the names of the ModelPartsExplanation objects for comparing.

        Returns
        -------
        list[float, list]

        """

        if isinstance(self.results[index_base], ModelPartsExplanation):
            if isinstance(index_base, int) and index_base < 0:
                index_base = self.results.index(self.results[index_base])

            def filter_objects(obj):
                if model_filter is not None and \
                        re.search(model_filter, obj.name) is None:
                    return False
                return True

            base_model = self[index_base]
            if base_model is None:
                raise Exception('Wrong index_base argument!')

            res = ModelPartsExplanationArray.__flatten(self.results)
            res.remove(self.results[index_base])

            res_filtered = [res[i] for i in range(len(res)) if filter_objects(res[i])]
            return base_model.compare(other=res_filtered, variable=variable, max_variables=max_variables,
                                      return_raw=return_raw)
        elif np.alltrue([isinstance(res, ModelPartsExplanationArray) for res in self.results]):
            return [
                res.compare(variable=variable, max_variables=max_variables, return_raw=return_raw,
                            index_base=index_base, model_filter=model_filter)
                for res in self.results
            ]
        else:
            raise Exception('Wrong result structure!')

    def plot_summary(self, model_filters: Optional[List[str]] = None, filter_labels: [List[str]] = None,
                     variables: Optional[List[str]] = None, max_variables: Optional[int] = None,
                     figsize: Optional[Tuple[int, int]] = None, index_base: Union[str, int] = -1,
                     significance_level: Optional[float] = None, fdr_correction: bool = True,
                     return_df: bool = False) -> None:
        """
        The function plots boxplots of comparison metrics of VI in the object if significance_level is provided.
        Otherwise, the results of the statistical test are plotted as barplots according to the significance_level.

        Parameters
        ----------
        variables : str, list[str], optional, default=None
            Variable for which the VI should be plotted. If None, the all columns is plotted.
        figsize : tuple(int, int), optional, default=(8, 8)
            Size of a figure.
        model_filters : list[str], optional, default=None
            List of regex expressions to filter the names of the ModelPartsExplanation objects for comparing.
            Each element in the list creates a new boxplot. If None, one boxplot / barplot of all results is plotted.
        filter_labels : list[str], optional, default=None
            Labels of model filters.
        index_base : int, str, default=-1
            Index of an explanation to be a base for comparisons.
        max_variables : int, optional, default=None
            Maximal number of variables from the current object to be taken into account.
        significance_level : float, optional, default=None
            A significance level of the statistical test (metric).
        fdr_correction : bool, default=True
            Add p-value correction for false discovery rate. Note that it is used only if significance_level is not None.
        return_df : bool, default=False
            If True, the method returns a dataframe on which a plot is created.

        """

        def extract_accepted_rejected(tab):
            n = len(tab)
            res = np.array(tab)
            reject = np.sum(res <= significance_level)
            accept = n - reject
            return accept / n, reject / n

        plt.subplots(figsize=figsize)

        if model_filters is None:
            results = self.compare(variable=variables, index_base=index_base,
                                   return_raw=True, max_variables=max_variables)

            results = ModelPartsExplanationArray.__flatten(results)

            if significance_level is not None and fdr_correction:
                _, results = fdrcorrection(results, significance_level)

            if filter_labels is not None:
                if len(filter_labels) == 1:
                    lbl = filter_labels
                else:
                    raise Exception('Incorrect length of filter_labels!')
            else:
                lbl = ['All values']

            if significance_level is None:
                plt.boxplot(results, labels=lbl, patch_artist=True)
            else:
                accepted, rejected = extract_accepted_rejected(results)

                plt.bar([0.8], accepted, 0.4, label='Accepted')
                plt.bar([1.2], rejected, 0.4, label='Rejected')
                plt.xticks([1], lbl)
                plt.legend()

        else:
            results = []
            for f in model_filters:
                tmp_out = self.compare(variable=variables, index_base=index_base,
                                       return_raw=True, model_filter=f, max_variables=max_variables)
                tmp_out = ModelPartsExplanationArray.__flatten(tmp_out)
                results.append(tmp_out)

            if filter_labels is not None:
                if len(filter_labels) == len(model_filters):
                    lbl = filter_labels
                else:
                    raise Exception('Incorrect length of filter_labels!')
            else:
                lbl = model_filters

            if significance_level is not None and fdr_correction:
                res_lengths = [len(r) for r in results]
                flat_results = [item for sublist in results for item in sublist]
                _, flat_results = fdrcorrection(flat_results, significance_level)
                q = 0
                for i in range(len(results)):
                    results[i] = flat_results[q:(q + res_lengths[i])]
                    q += res_lengths[i]

            if significance_level is None:
                plt.boxplot(results, labels=lbl, patch_artist=True)
            else:
                accepted = []
                rejected = []

                for r in results:
                    acc, rej = extract_accepted_rejected(r)
                    accepted.append(acc)
                    rejected.append(rej)

                bar_width = 0.4
                x = lbl
                x_axis = np.arange(len(x))

                plt.bar(x_axis, accepted, bar_width, label='Accepted')
                plt.bar(x_axis + bar_width, rejected, bar_width, label='Rejected')
                plt.legend()

                plt.xticks(x_axis + bar_width / 2, x)

        if significance_level is None:
            plt.ylabel(r'p-values')
            plt.title(f'Summary of {self.explanation_type} for {self.name}')
            plt.xlabel('Filter')
        else:
            plt.ylabel('Fraction of tests')
            plt.title(
                f'Summary of {self.explanation_type} for {self.name} with p-value significance level {significance_level}')
            plt.xlabel('Filter')

        if return_df:
            return results

    @staticmethod
    def __flatten(lst):
        out = []
        for i in range(len(lst)):
            if not (isinstance(lst[i], list) or isinstance(lst[i], ModelPartsExplanationArray)):
                out.append(lst[i])
            else:
                tmp = ModelPartsExplanationArray.__flatten(lst[i])
                out = out + tmp
        return out

    def __str__(self) -> str:
        return f"ModelPartsExplanationArray {self.name}"

    def __repr__(self) -> str:
        return f"<ModelPartsExplanationArray {self.name}>"



from __future__ import annotations

import numpy as np
import warnings
import pandas as pd
import xgboost as xgb

from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Optional, Protocol, Any, Dict, List, Callable
from sklearn.ensemble import RandomForestClassifier
from sklearn.preprocessing import OrdinalEncoder
from sklearn.metrics import accuracy_score, f1_score, balanced_accuracy_score, precision_score
from sklearn.metrics import roc_auc_score, recall_score
from sklearn.utils.validation import check_is_fitted
from imblearn.metrics import geometric_mean_score, specificity_score
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.model_selection import RandomizedSearchCV as RS
from sklearn.model_selection import GridSearchCV as GS

from edgaro.data.dataset import Dataset
from edgaro.base.base_transformer import BaseTransformer
from edgaro.base.utils import print_unbuffered


class Model(BaseTransformer, ABC):
    """
    The abstract class to define a Machine Learning model for a single Dataset.

    Parameters
    ----------
    name : str
        A name of the Model.
    test_size : float, optional, default=None
        Test size for a Dataset object in case it was not train-test-split. If a Dataset object was not train-test-split
        and the parameter has value None, the training will be done on the all data.
    random_state : int, optional, default=None
        Random state seed.
    verbose : bool, default=False
        Print messages during calculations.

    Attributes
    ----------
    name : str
        A name of the Model.
    test_size : float, optional
        Test size for a Dataset object in case it was not train-test-split. If a Dataset object was not train-test-split
        and the parameter has value None, the training will be done on the all data.
    random_state : int, optional
        Random state seed.
    verbose : bool
        Print messages during calculations.
    majority_class_label : str, optional, default=None
        The label of the majority class. It is recommended to pass this argument; otherwise, it will be guessed.
        The guess may be wrong if the dataset is balanced - consequently, experiment results may be wrong. If None,
        it will be tried to extract the information from the majority_class_label attribute of the Dataset object.

    """

    def __init__(self, name: str = '', test_size: Optional[float] = None, random_state: Optional[int] = None,
                 verbose: bool = False, majority_class_label: Optional[str] = None) -> None:
        super().__init__()
        self.__transform_to_probabilities = False
        self.__train_dataset = None
        self.__test_dataset = None
        self.test_size = test_size
        self.name = name
        self.__label_encoders = {}
        self.__target_encoder = None
        self.random_state = random_state
        self.__was_fitted = False
        self.verbose = verbose
        self.majority_class_label = majority_class_label

    def fit(self, dataset: Dataset, print_scores: bool = False) -> None:
        """
        Fit the Model.

        The fitting process includes encoding the categorical variables with OrdinalEncoder (from scikit-learn library)
        and target encoding (custom encoding, the minority class is encoded as 1, the majority class as 0).

        The method assumes that categorical variables, which has to be encoded,
        are one of the types: 'category', 'object'.

        Parameters
        ----------
        dataset : Dataset
            The object to fit Model on.
        print_scores : bool, default=False
            Indicates whether model evaluation on a test dataset should be printed at the end of fitting.

        """
        if self.verbose:
            print_unbuffered(f'Model {self.__repr__()} is being fitted with {dataset.name}')

        if self.majority_class_label is None and dataset.majority_class_label is not None:
            self.majority_class_label = dataset.majority_class_label

        if dataset.data is None or (dataset.data is not None and len(dataset.data) == 0):
            raise Exception('The dataset has empty data!')
        elif dataset.target is None or (dataset.target is not None and len(dataset.target) == 0):
            raise Exception('The dataset has empty data!')
        else:
            if not dataset.check_binary_classification():
                raise Exception('Dataset does not represent binary classification task!')

            if not dataset.was_split and self.test_size is None:
                warnings.warn(
                    'Dataset was not train-test-split! The training dataset will be the same as the test dataset.')
                self.__train_dataset = deepcopy(dataset)
                self.__test_dataset = deepcopy(dataset)
            else:
                if dataset.was_split and self.test_size is not None:
                    warnings.warn('Dataset was train-test-split! Dataset will not be split the second time.')
                elif not dataset.was_split and self.test_size is not None:
                    dataset.train_test_split(test_size=self.test_size, random_state=self.random_state)
                self.__train_dataset = deepcopy(dataset.train)
                self.__test_dataset = deepcopy(dataset.test)

            ds = deepcopy(self.__train_dataset)

            columns_to_encode = list(dataset.data.select_dtypes(include=['category', 'object']))
            for col in columns_to_encode:
                le = OrdinalEncoder(handle_unknown='use_encoded_value', unknown_value=-1)
                ds.data[col] = le.fit_transform(ds.data[[col]])
                self.__label_encoders[col] = le

            self.__target_encoder = _TargetEncode(majority_class_label=self.majority_class_label)
            ds.target = self.__target_encoder.fit_transform(ds.target)
            self.__train_dataset.target = self.__target_encoder.transform(self.__train_dataset.target)

            if self.name == '':
                self.name = dataset.name

            self._fit(ds)

            if print_scores:
                print_unbuffered(self.evaluate())

            self.__was_fitted = True

            if self.verbose:
                print_unbuffered(f'Model {self.__repr__()} was fitted with {dataset.name}')

    @abstractmethod
    def _fit(self, dataset: Dataset) -> None:
        pass

    @property
    def was_fitted(self) -> bool:
        """
        The information whether the Model was fitted.

        Returns
        -------
        bool
        """
        return self.__was_fitted

    def transform_data(self, dataset: Dataset) -> Dataset:
        """
        Encode dataset.data with the rules generated after fitting this object.

        Parameters
        ----------
        dataset : Dataset
            A Dataset object, where `.data` attribute will be encoded. The method returns a new object.

        Returns
        -------
        Dataset

        """
        if dataset.data is None or (dataset.data is not None and len(dataset.data) == 0):
            raise Exception('The dataset has empty data!')
        else:
            df = Dataset(
                name=dataset.name,
                dataframe=deepcopy(dataset.data),
                target=deepcopy(dataset.target)
            )
            if len(self.__label_encoders) > 0:
                for key, le in self.__label_encoders.items():
                    df.data[key] = le.transform(pd.DataFrame(df.data[key]))
            return df

    def transform_target(self, dataset: Dataset) -> Dataset:
        """
        Encode dataset.target with the rules generated after fitting this object.

        Parameters
        ----------
        dataset : Dataset
            A Dataset object, where `.target` attribute will be encoded. The method returns a new object.

        Returns
        -------
        Dataset

        """
        if dataset.target is None or (dataset.target is not None and len(dataset.target) == 0):
            raise Exception('The dataset has empty data!')
        else:
            df = Dataset(
                name=dataset.name,
                dataframe=deepcopy(dataset.data),
                target=deepcopy(dataset.target)
            )
            if self.__target_encoder is not None:
                df.target = self.__target_encoder.transform(df.target)
            return df

    def predict(self, dataset: Dataset) -> Dataset:
        """
        Predict the class for a Dataset object.

        Parameters
        ----------
        dataset : Dataset
            A Dataset object to make predictions on.

        Returns
        -------
        Dataset

        """
        if dataset.data is None:
            raise Exception('Data in dataset is not provided!')

        df = self.transform_data(dataset)
        model_name = '_' + self.name if not self.name == dataset.name else ''
        name = dataset.name + model_name + '_predicted'

        if self.verbose:
            print_unbuffered(f'Model {self.__repr__()} predicted on {dataset.name}')

        return self._predict(df, output_name=name)

    @abstractmethod
    def _predict(self, dataset: Dataset, output_name: str) -> Dataset:
        pass

    def predict_proba(self, dataset: Dataset) -> Dataset:
        """
        Predict the probability of class `1` for a Dataset object.

        Parameters
        ----------
        dataset : Dataset
            A Dataset object to make predictions on.

        Returns
        -------
        Dataset

        """
        if dataset.data is None:
            raise Exception('Data in dataset is not provided!')

        df = self.transform_data(dataset)
        model_name = '_' + self.name if not self.name == dataset.name else ''
        name = dataset.name + model_name + '_predicted_probabilities'

        if self.verbose:
            print_unbuffered(f'Model {self.__repr__()} predicted probabilities on {dataset.name}')

        return self._predict_proba(df, output_name=name)

    @abstractmethod
    def _predict_proba(self, dataset: Dataset, output_name: str) -> Dataset:
        pass

    def set_params(self, **params) -> None:
        """
        Set params for Model.

        Parameters
        ----------
        params : dict
            The parameters to be set.

        """
        if 'test_size_model' in params.keys():
            self.test_size = params.pop('test_size_model')
        self._set_params(**params)

    @abstractmethod
    def _set_params(self, **params) -> None:
        pass

    @abstractmethod
    def get_params(self) -> Dict:
        """
        Get parameters of Model.

        Returns
        -------
        Dict, list
            The parameters.

        """
        pass

    def transform(self, dataset: Dataset) -> Dataset:
        """
        A function to make the Model compatible with BaseTransformer.

        It can either return predicted classes or predicted probabilities - it can be set using
        `set_transform_to_probabilities` and `set_transform_to_classes` functions.

        Parameters
        ----------
        dataset : Dataset
            A Dataset object to be transformed.
        Returns
        -------
        Dataset
        """
        return self.__transform(dataset)

    def __transform(self, dataset: Dataset) -> Dataset:
        if self.__transform_to_probabilities:
            return self.predict_proba(dataset)
        else:
            return self.predict(dataset)

    def set_transform_to_probabilities(self) -> None:
        """
        Make `transform` function return probabilities.
        """
        self.__transform_to_probabilities = True

    def set_transform_to_classes(self) -> None:
        """
        Make `transform` function return classes..
        """
        self.__transform_to_probabilities = False

    def get_train_dataset(self) -> Optional[Dataset]:
        """
        Get a Dataset used for a training process.

        Returns
        -------
        Dataset

        """
        return self.__train_dataset

    def get_test_dataset(self) -> Optional[Dataset]:
        """
        Get a Dataset used for a test process.

        Returns
        -------
        Dataset

        """
        return self.__test_dataset

    def get_category_colnames(self) -> List[str]:
        """
        Get category column names, which were encoded during the fitting process.

        Returns
        -------
        list(str)
        """
        if len(self.__label_encoders) > 0:
            return list(self.__label_encoders.keys())
        else:
            return []

    def evaluate(self, metrics_output_class: Optional[List[Callable[[pd.Series, pd.Series], float]]] = None,
                 metrics_output_probabilities: Optional[List[Callable[[pd.Series, pd.Series], float]]] = None,
                 ds: Optional[Dataset] = None) -> pd.DataFrame:
        """
        Evaluate model.

        Parameters
        ----------
        metrics_output_class : list[Callable[[pd.Series, pd.Series], float]], optional, default=None
            List of functions to calculate metrics on predicted classes. If None is passed, accuracy, balanced accuracy,
            precision, recall, specificity, f1, f1_weighted, geometric mean score are used.
        metrics_output_probabilities : list[Callable[[pd.Series, pd.Series], float]], optional, default=None
            List of functions to calculate metrics on predicted probabilities. If None is passed, ROC AUC is used.
        ds : Dataset, optional, default=None
            A Dataset object to calculate metric on. If None is passed, test Dataset from fitting is used.

        Returns
        -------
        pd.DataFrame
        """

        if self.verbose:
            print_unbuffered(f'Model {self.__repr__()} is being evaluated')

        results = {}
        if ds is None:
            if self.__test_dataset is not None:
                ds = self.__test_dataset
            else:
                raise Exception('There is not test dataset and the ds argument was not provided!')
        if metrics_output_class is None and metrics_output_probabilities is None:
            def f1_weighted(y_true, y_pred):
                return f1_score(y_true, y_pred, average='weighted')

            metrics_output_class = [accuracy_score, balanced_accuracy_score, precision_score,
                                    recall_score, specificity_score, f1_score, f1_weighted, geometric_mean_score]
            metrics_output_probabilities = [roc_auc_score]
        if metrics_output_class is not None and len(metrics_output_class) > 0:
            y_hat = self.predict(ds)
            for f in metrics_output_class:
                results[f.__name__] = f(self.__target_encoder.transform(ds.target), y_hat.target)
        if metrics_output_probabilities is not None and len(metrics_output_probabilities) > 0:
            y_hat = self.predict_proba(ds)
            for f in metrics_output_probabilities:
                results[f.__name__] = f(self.__target_encoder.transform(ds.target), y_hat.target)

        if self.verbose:
            print_unbuffered(f'Model {self.__repr__()} was evaluated')

        return pd.DataFrame(results.items(), columns=['metric', 'value'])

    def __str__(self) -> str:
        return f"{self.__class__.__name__} model with name {self.name}"

    def __repr__(self) -> str:
        return f"<{self.name} {self.__class__.__name__} model>"


class _TargetEncode(BaseEstimator, TransformerMixin):
    def __init__(self, majority_class_label: Optional[str] = None) -> None:
        self.mapping = None
        super().__init__()
        self.majority_class_label = majority_class_label

    def fit(self, y: pd.Series) -> _TargetEncode:
        names, counts = np.unique(y, return_counts=True)
        if names is None:
            raise Exception('The input vector is wrong!')
        elif len(names) != 2:
            raise Exception('The input vector do not has two classes!')
        elif self.majority_class_label is not None:
            names = names.tolist()
            ind = names.index(self.majority_class_label)

            if (ind == 1 and counts[0] > counts[1]) or (ind == 0 and counts[1] > counts[0]):
                print('Majority class label may be wrong!')
                print('It may be an error or a result of undersampling method ' +
                      '(it might have happened that there is less observations in the initial majority class ' +
                      'then in the initial minority class)!')

            self.mapping = {
                names[0 if ind == 1 else 1]: 1,
                names[ind]: 0
                }
            return self

        if counts[0] < counts[1]:
            index_min = 0
            index_max = 1
        else:
            index_min = 1
            index_max = 0

        self.mapping = {
            names[index_min]: 1,
            names[index_max]: 0
        }
        return self

    def transform(self, y: pd.Series) -> pd.Series:
        return y.map(self.mapping)


class SKLEARNModelProtocol(Protocol):
    """
    A Protocol to define the expected structure of a Model from `scikit-learn` library.
    """

    def fit(self, X: pd.DataFrame, y: pd.Series) -> Any:
        ...

    def predict(self, X: pd.DataFrame) -> np.ndarray:
        ...

    def predict_proba(self, X: pd.DataFrame) -> np.ndarray:
        ...

    def get_params(self) -> Dict:
        ...

    def set_params(self, **params) -> Any:
        ...


class ModelFromSKLEARN(Model):
    """
    Create Model from a model in `scikit-learn` library.

    Parameters
    ----------
    base_model : SKLEARNModelProtocol
        A model from `scikit-learn` library. Note: this object has to be clean (not fitted).
    name : str
        A name of the Model.
    test_size : float, optional, default=None
        Test size for a Dataset object in case it was not train-test-split. If a Dataset object was not train-test-split
        and the parameter has value None, the training will be done on the all data.
    random_state : int, optional, default=None
        Random state seed.
    verbose : bool, default=False
        Print messages during calculations.

    """

    def __init__(self, base_model: SKLEARNModelProtocol, name: str = '', test_size: Optional[float] = None,
                 random_state: Optional[int] = None, verbose: bool = False) -> None:
        super().__init__(name=name, test_size=test_size, random_state=random_state)
        self._model = base_model
        self.verbose = verbose

    def _fit(self, dataset: Dataset) -> None:
        self._model.set_params(verbose=self.verbose)

        if 'random_state' in self._model.get_params().keys() and \
                self._model.get_params()['random_state'] is None and \
                self.random_state is not None:
            self._model.set_params(**{'random_state': self.random_state})

        self._model.fit(dataset.data, dataset.target)

    @property
    def was_fitted(self) -> bool:
        try:
            check_is_fitted(self._model)
            return True
        except (Exception,):
            return False

    def _predict(self, dataset: Dataset, output_name: str) -> Dataset:
        return Dataset(
            name=output_name,
            dataframe=None,
            target=pd.Series(self._model.predict(dataset.data))
        )

    def _predict_proba(self, dataset: Dataset, output_name: str) -> Dataset:
        return Dataset(
            name=output_name,
            dataframe=None,
            target=pd.Series(self._model.predict_proba(dataset.data)[:, 1])
        )

    def _set_params(self, **params) -> None:
        return self._model.set_params(**params)

    def get_params(self) -> Dict:
        return self._model.get_params()


class RandomForest(ModelFromSKLEARN):
    """
    Create RandomForest Model from a RandomForestClassifier implementation in `scikit-learn` library.

    Parameters
    ----------
    name : str
        A name of the Model.
    test_size : float, optional, default=None
        Test size for a Dataset object in case it was not train-test-split. If a Dataset object was not train-test-split
        and the parameter has value None, the training will be done on the all data.
    random_state : int, optional, default=None
        Random state seed.
    verbose : bool, default=False
        Print messages during calculations.
    *args : tuple, optional
        Additional parameters for RandomForestClassifier from `scikit-learn` library.
    **kwargs : dict, optional
        Additional parameters for RandomForestClassifier from `scikit-learn` library.

    """

    def __init__(self, name: str = '', test_size: Optional[float] = None, random_state: Optional[int] = None,
                 verbose: bool = False, *args, **kwargs) -> None:
        super().__init__(RandomForestClassifier(*args, **kwargs), name=name, test_size=test_size,
                         random_state=random_state, verbose=verbose)


class XGBoost(ModelFromSKLEARN):
    """
    Create XGBoost Model from a XGBClassifier implementation in `xgboost` library.

    Parameters
    ----------
    name : str
        A name of the Model.
    test_size : float, optional, default=None
        Test size for a Dataset object in case it was not train-test-split. If a Dataset object was not train-test-split
        and the parameter has value None, the training will be done on the all data.
    random_state : int, optional, default=None
        Random state seed.
    verbose : bool, default=False
        Print messages during calculations.
    *args : tuple, optional
        Additional parameters for XGBClassifier from `xgboost` library.
    **kwargs : dict, optional
        Additional parameters for XGBClassifier from `xgboost` library.

    """

    def __init__(self, name: str = '', test_size: Optional[float] = None, random_state: Optional[int] = None,
                 verbose: bool = False, *args, **kwargs) -> None:
        super().__init__(
            xgb.XGBClassifier(eval_metric='logloss' if 'eval_metric' not in kwargs.keys() else kwargs['eval_metric'],
                              *args, **kwargs),
            name=name, test_size=test_size, random_state=random_state, verbose=verbose
        )


class RandomSearchCV(ModelFromSKLEARN):
    """
    Create Model to perform Random Search on any of the model implementation matching SKLEARNModelProtocol.

    Parameters
    ----------
    base_model : SKLEARNModelProtocol
        A model from `scikit-learn` library. Note: this object has to be clean (not fitted).
    param_grid : Dict
        A parameter grid for searching.
    n_iter : int
        Number of iterations to be performed.
    cv : int
        Number of cross-validation folds.
    scoring : str
        Name of a function to be used to choose the best model.
    name : str
        A name of the Model.
    test_size : float, optional, default=None
        Test size for a Dataset object in case it was not train-test-split. If a Dataset object was not train-test-split
        and the parameter has value None, the training will be done on the all data.
    random_state : int, optional, default=None
        Random state seed.
    verbose : bool, default=False
        Print messages during calculations.

    """

    def __init__(self, base_model: ModelFromSKLEARN, param_grid: Dict, n_iter: int = 10, cv: int = 5,
                 scoring: str = 'balanced_accuracy', name: str = '', test_size: Optional[float] = None,
                 random_state: Optional[int] = None, verbose: bool = False, *args, **kwargs) -> None:
        super().__init__(
            RS(base_model._model, param_grid, cv=cv, scoring=scoring, n_iter=n_iter, *args, **kwargs),
            name=name,
            test_size=test_size,
            random_state=random_state,
            verbose=verbose
        )


class GridSearchCV(ModelFromSKLEARN):
    """
    Create Model to perform Grid Search on any of the model implementation matching SKLEARNModelProtocol.

    Parameters
    ----------
    base_model : SKLEARNModelProtocol
        A model from `scikit-learn` library. Note: this object has to be clean (not fitted).
    param_grid : Dict
        A parameter grid for searching.
    cv : int
        Number of cross-validation folds.
    scoring : str
        Name of a function to be used to choose the best model.
    name : str
        A name of the Model.
    test_size : float, optional, default=None
        Test size for a Dataset object in case it was not train-test-split. If a Dataset object was not train-test-split
        and the parameter has value None, the training will be done on the all data.
    random_state : int, optional, default=None
        Random state seed.
    verbose : bool, default=False
        Print messages during calculations.

    """

    def __init__(self, base_model: ModelFromSKLEARN, param_grid: Dict, cv: int = 5,
                 scoring: str = 'balanced_accuracy', name: str = '', test_size: Optional[float] = None,
                 random_state: Optional[int] = None, verbose: bool = False, *args, **kwargs) -> None:
        if 'random_state' in base_model._model.get_params().keys():
            base_model._model.set_params(**{'random_state': random_state})

        super().__init__(
            GS(base_model._model, param_grid, cv=cv, scoring=scoring, *args, **kwargs),
            name=name,
            test_size=test_size,
            random_state=random_state,
            verbose=verbose
        )



import pytest
import re
import openml

from copy import deepcopy

from edgaro.data.dataset_array import DatasetArray, DatasetArrayFromOpenMLSuite, DatasetArrayFromDirectory
from edgaro.data.dataset_array import load_benchmarking_set
from edgaro.data.dataset import Dataset, DatasetFromOpenML

from .resources.objects import *


def test_dataset_array_creation():
    try:
        ds1, ds2 = Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)
        DatasetArray([ds1, ds2])
    except (Exception,):
        assert False


@pytest.mark.parametrize('ds1, ds2', [
    (Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)),
    (DatasetFromOpenML(task_id=task_id_1, apikey=APIKEY), DatasetFromOpenML(task_id=task_id_2, apikey=APIKEY)),
    (Dataset(name_3, df_3, target_3), Dataset(name_1, df_1, target_1))
])
class TestDatasetArrayBasicProperties:
    def test_content_get_item(self, ds1, ds2):
        tab = DatasetArray([deepcopy(ds1), deepcopy(ds2)])
        assert len(tab) == 2
        assert tab[0] == ds1
        assert tab[1] == ds2
        assert tab[ds1.name] == ds1
        assert tab[ds2.name] == ds2

        assert len(tab[[ds1.name, ds2.name]]) == 2

        tmp = deepcopy(tab)
        tmp.name += '_subset'
        assert tab[[ds1.name, ds2.name]] == tmp

    def test_str_repr(self, ds1, ds2):
        tab = DatasetArray([deepcopy(ds1), deepcopy(ds2)])
        try:
            str(tab)
            repr(tab)
        except (Exception,):
            assert False

    def test_verbose(self, ds1, ds2, capsys):
        ds = DatasetArray([deepcopy(ds1), deepcopy(ds2)], verbose=True)
        captured = capsys.readouterr()
        assert captured.out.startswith(f'DatasetArray {ds.__repr__()} created\n')

    def test_dataset_array_with_dataset_arrays(self, ds1, ds2):
        tab = DatasetArray([
            deepcopy(ds1),
            DatasetArray([deepcopy(ds2), deepcopy(ds1)], name='example')
        ])

        assert len(tab) == 2
        assert tab[0] == ds1
        assert tab[1][0] == ds2
        assert tab[1][1] == ds1
        assert tab[ds1.name] == ds1
        assert tab['example'][0] == ds2
        assert tab['example'][1] == ds1

    def test_datasetarray_get_out_index(self, ds1, ds2):
        tab = DatasetArray([
            deepcopy(ds1),
            DatasetArray([deepcopy(ds2), deepcopy(ds1)], name='example')
        ])

        assert tab[['non_exist_1', 'non_exist_2']] is None

    def test_datasetarray_get_wrong_key(self, ds1, ds2):
        tab = DatasetArray([
            deepcopy(ds1),
            DatasetArray([deepcopy(ds2), deepcopy(ds1)], name='example')
        ])

        assert tab[[100, 101]] is None


def test_benchmarking_set():
    try:
        df = load_benchmarking_set(apikey=APIKEY)
    except (Exception,):
        assert False

    assert isinstance(df, DatasetArray)


class TestUniqueNames:
    @pytest.mark.parametrize('ds1, ds2', [
        (Dataset(name_1, df_1, target_1), Dataset(name_1, df_2, target_2)),
        (Dataset(name_2, df_1, target_1), Dataset(name_2, df_2, target_2)),
        (Dataset(name_2, df_1, target_1), DatasetArray([Dataset(name_2, df_2, target_2), Dataset(name_1, df_2, target_2)], name=name_2))
    ])
    def test_no_unique_names_1(self, ds1, ds2):
        with pytest.raises(Exception):
            DatasetArray([ds1, ds2])

    def test_no_unique_names_2(self):
        with pytest.raises(Exception):
            (Dataset(name_2, df_1, target_1), DatasetArray([Dataset(name_1, df_2, target_2), Dataset(name_1, df_2, target_2)], name=name_1))


class TestEqual:
    @pytest.mark.parametrize('ds1, ds2', [
        (Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)),
        (DatasetFromOpenML(task_id=task_id_1, apikey=APIKEY), DatasetFromOpenML(task_id=task_id_2, apikey=APIKEY)),
        (Dataset(name_3, df_3, target_3), Dataset(name_2, df_3, target_3)),
        (Dataset(name_3, df_3, target_3), DatasetArray([Dataset(name_2, df_3, target_3), Dataset(name_1, df_3, target_3)]))
    ])
    def test_equal(self, ds1, ds2):
        tab1 = DatasetArray([deepcopy(ds1), deepcopy(ds2)])
        tab2 = DatasetArray([deepcopy(ds1), deepcopy(ds2)])
        assert tab1 == tab2

    @pytest.mark.parametrize('ds1,ds2', [
        (Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)),
        (DatasetFromOpenML(task_id=task_id_1, apikey=APIKEY), DatasetFromOpenML(task_id=task_id_2, apikey=APIKEY)),
        (DatasetArray([Dataset(name_2, df_3, target_3), Dataset(name_1, df_3, target_3)]), Dataset(name_3, df_3, target_3)),
        (DatasetArray([Dataset(name_1, df_3, target_3), Dataset(name_2, df_3, target_3)]), Dataset(name_3, df_3, target_3))
    ])
    def test_equal_reverse(self, ds1, ds2):
        tab1 = DatasetArray([deepcopy(ds1), deepcopy(ds2)])
        tab2 = DatasetArray([deepcopy(ds2), deepcopy(ds1)])
        assert tab1 == tab2

    @pytest.mark.parametrize('ds1,ds2', [
        (Dataset(name_1, df_1, target_1), Dataset(name_1, df_1, target_2)),
        (Dataset(name_1, df_1, target_1), Dataset(name_1, df_2, target_1)),
        (Dataset(name_1, df_1, target_1), Dataset(name_2, df_1, target_1)),
        (
                DatasetArray([
                    DatasetArray([Dataset(name_1, df_2, target_1), Dataset(name_2, df_3, target_3)]),
                    Dataset(name_2, df_3, target_3)
                ]),
                DatasetArray([
                    DatasetArray([Dataset(name_1, df_2, target_2), Dataset(name_2, df_3, target_3)]),
                    Dataset(name_2, df_3, target_3)
                ])
        )
    ])
    def test_not_equal(self, ds1, ds2):
        tab1 = DatasetArray([deepcopy(ds1)])
        tab2 = DatasetArray([deepcopy(ds2)])
        assert not tab1 == tab2

    @pytest.mark.parametrize('ds1,ds2', [
        (Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)),
        (DatasetFromOpenML(task_id=task_id_1, apikey=APIKEY), DatasetFromOpenML(task_id=task_id_2, apikey=APIKEY))
    ])
    def test_not_equal_name(self, ds1, ds2):
        tab1 = DatasetArray([deepcopy(ds1), deepcopy(ds2)], name=ds1.name)
        tab2 = DatasetArray([deepcopy(ds2), deepcopy(ds1)], name=ds2.name)
        assert not tab1 == tab2

    @pytest.mark.parametrize('da1,da2', [
        (
            [Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)],
            [Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_1)]
        ),
        (
            [Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)],
            [Dataset(name_1, df_1, target_1), Dataset(name_2, df_1, target_2)]
        ),
        (
            [Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)],
            [Dataset(name_1, df_1, target_1), Dataset(name_1 + '_extra', df_2, target_2)],
        ),
        # reverse order
        (
            [Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)],
            [Dataset(name_1, df_2, target_2), Dataset(name_1 + '_extra', df_1, target_1)],
        ),
        (
            [Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)],
            [Dataset(name_2, df_1, target_2), Dataset(name_1, df_1, target_1)]
        ),
        (
            [Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2), Dataset(name_1 + '_extra', df_1, target_1)],
            [Dataset(name_2, df_1, target_2), Dataset(name_1, df_1, target_1)]
        ),
        (
            [Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2), DatasetArray([Dataset(name_2, df_3, target_3), Dataset(name_1, df_3, target_3)], name='xx')],
            [Dataset(name_2, df_1, target_2), Dataset(name_1, df_1, target_1), DatasetArray([Dataset(name_2, df_3, target_3), Dataset(name_1, df_3, target_3)], name='xxyy')]
        )
    ])
    def test_not_equal_2(self, da1, da2):
        tab1 = DatasetArray(da1)
        tab2 = DatasetArray(da2)
        assert not tab1 == tab2


class TestOtherFunctions:
    @pytest.mark.parametrize('ds', [
        Dataset(name_1, df_1, target_1),
        DatasetArray([Dataset(name_2, df_2, target_2), Dataset(name_1, df_2, target_2)], name=name_2),
        DatasetArray([
            Dataset(name_2, df_2, target_2),
            Dataset(name_1, df_2, target_2),
            DatasetArray([Dataset(name_2, df_2, target_2), Dataset(name_1, df_2, target_2)], name=name_2 + 'x')
        ], name=name_2),
        Dataset(name_1, None, target_1),
        Dataset(name_1, df_1, None),
        Dataset(name_1, None, None)
    ])
    def test_str_repr(self, ds):
        try:
            str(ds)
            repr(ds)
        except (Exception,):
            assert False

    def test_remove_nans(self):
        da = DatasetArray([
            Dataset(name_4_nans + '_0', df_4_nans, target_4_nans),
            Dataset(name_4_nans + '_1', df_4_nans, target_4_nans)
        ])
        da = deepcopy(da)
        da.remove_nans()

        assert da[0].data.shape == (1, 2)
        assert da[0].target.shape == (1,)
        assert da[1].data.shape == (1, 2)
        assert da[1].target.shape == (1,)

    def test_remove_nans_2(self):
        da = DatasetArray([
            Dataset(name_4_nans + '_0', df_4_nans, target_4_nans),
            Dataset(name_4_nans + '_1', df_4_nans, target_4_nans),
            DatasetArray([
                Dataset(name_4_nans + '_0', df_4_nans, target_4_nans),
                Dataset(name_4_nans + '_1', df_4_nans, target_4_nans),
            ])
        ])
        da = deepcopy(da)
        da.remove_nans()

        assert da[0].data.shape == (1, 2)
        assert da[0].target.shape == (1,)
        assert da[1].data.shape == (1, 2)
        assert da[1].target.shape == (1,)

        assert len(da[2]) == 2
        assert da[2][0].data.shape == (1, 2)
        assert da[2][0].target.shape == (1,)
        assert da[2][1].data.shape == (1, 2)
        assert da[2][1].target.shape == (1,)

    @pytest.mark.parametrize('ds1, ds2', [
        (DatasetFromOpenML(task_id=task_id_1, apikey=APIKEY), DatasetFromOpenML(task_id=task_id_2, apikey=APIKEY))
    ])
    def test_train_test_split(self, ds1, ds2):
        tab = DatasetArray([deepcopy(ds1), deepcopy(ds2)])

        try:
            tab.train_test_split()
        except (Exception,):
            assert False

        for ds in tab:
            assert ds.was_split

    @pytest.mark.parametrize('ds1, ds2', [
        (DatasetFromOpenML(task_id=task_id_1, apikey=APIKEY), DatasetFromOpenML(task_id=task_id_2, apikey=APIKEY))
    ])
    def test_train_test_split_verbose(self, ds1, ds2, capsys):
        tab = DatasetArray([deepcopy(ds1), deepcopy(ds2)], verbose=True)
        tab.train_test_split()

        captured = capsys.readouterr()
        assert captured.out.endswith(f'DatasetArray {tab.__repr__()} was train-test-split\n')

    @pytest.mark.parametrize('ds1, ds2', [
        (DatasetFromOpenML(task_id=task_id_1, apikey=APIKEY), DatasetFromOpenML(task_id=task_id_2, apikey=APIKEY))
    ])
    def test_train_test_split_train_property(self, ds1, ds2):
        tab = DatasetArray([deepcopy(ds1), deepcopy(ds2)], verbose=True)
        tab.train_test_split()

        assert tab.train.name.endswith('_train')

    @pytest.mark.parametrize('ds1, ds2', [
        (DatasetFromOpenML(task_id=task_id_1, apikey=APIKEY), DatasetFromOpenML(task_id=task_id_2, apikey=APIKEY))
    ])
    def test_train_test_split_test_property(self, ds1, ds2):
        tab = DatasetArray([deepcopy(ds1), deepcopy(ds2)], verbose=True)
        tab.train_test_split()

        assert tab.test.name.endswith('_test')

    @pytest.mark.parametrize('ds1, ds2', [
        (DatasetFromOpenML(task_id=task_id_1, apikey=APIKEY), DatasetFromOpenML(task_id=task_id_2, apikey=APIKEY))
    ])
    def test_remove_outliers(self, ds1, ds2):
        tab = DatasetArray([deepcopy(ds1), deepcopy(ds2)], verbose=True)

        try:
            tab.remove_outliers()
        except (Exception,):
            assert False

    @pytest.mark.parametrize('ds1, ds2', [
        (DatasetFromOpenML(task_id=task_id_1, apikey=APIKEY), DatasetFromOpenML(task_id=task_id_2, apikey=APIKEY))
    ])
    def test_head(self, ds1, ds2):
        ds = DatasetArray([ds1, ds2])
        ds_head = ds.head(2)

        for d in ds_head:
            assert d.data.shape[0] == 2
            assert d.target.shape[0] == 2


@pytest.fixture(
    scope='module',
    params=[
        suite_name_1,
        # suite_name_2 # commented because of waiting time
    ]
)
def dataset_openml(request):
    return DatasetArrayFromOpenMLSuite(request.param, apikey=APIKEY)


@pytest.mark.parametrize('task', [
    suite_name_1
])
def test_apikey_fail_none(task):
    tmp_code = openml.config.apikey
    openml.config.apikey = ''

    with pytest.raises(Exception):
        DatasetArrayFromOpenMLSuite(task, apikey=None)

    openml.config.apikey = tmp_code


@pytest.mark.parametrize('task', [
    suite_name_1
])
def test_apikey_with_apikey_empty_openml(task):
    tmp_code = openml.config.apikey
    openml.config.apikey = ''

    try:
        DatasetArrayFromOpenMLSuite(task, apikey=APIKEY)
    except (Exception,):
        assert False
    finally:
        openml.config.apikey = tmp_code


@pytest.mark.parametrize('task', [
    suite_name_1
])
def test_apikey_verbose(task, capsys):
    DatasetArrayFromOpenMLSuite(task, apikey=APIKEY, verbose=True)

    captured = capsys.readouterr()
    assert f'Benchmark suite data was downloaded for {task}\n' in captured.out
    assert f'DatasetArray from OpenML benchmark suite {task} was created' in captured.out


class TestOpenMLSuite:
    def test_print_description_from_openml_suite(self, dataset_openml):
        ds = deepcopy(dataset_openml)
        out = ds.openml_description()
        pattern = '^Name: .* Description: .*$'
        assert re.match(pattern, out.replace('\n', ' '))

    def test_remove_non_binary_target_datasets(self, dataset_openml):
        # These OpenML suites contain also non-binary target datasets
        da = deepcopy(dataset_openml)
        length_1 = len(da)

        da.remove_non_binary_target_datasets()
        length_2 = len(da)

        assert length_2 < length_1

    def test_remove_non_binary_target_verbose(self, dataset_openml, capsys):
        # These OpenML suites contain also non-binary target datasets
        da = deepcopy(dataset_openml)
        da.verbose = True
        da.remove_non_binary_target_datasets()

        captured = capsys.readouterr()
        assert captured.out.startswith(f'Non binary datasets were removed from DatasetArray {da.__repr__()}\n')


class TestRemoveNonBinary:
    @pytest.mark.parametrize('ds,expected_binary_n', [
        ([Dataset(name_1, df_1, target_1), Dataset(name_2, df_1, target_1_fake), Dataset(name_3, df_1, target_1_fake)], 1),
        ([Dataset(name_1, df_1, target_1), Dataset(name_2, df_1, target_1), Dataset(name_3, df_1, target_1_fake)], 2),
        ([Dataset(name_1, df_1, target_1), Dataset(name_2, df_1, target_1), Dataset(name_3, df_1, target_1)], 3),
        ([Dataset(name_1, df_1, target_1_fake), Dataset(name_2, df_1, target_1_fake), Dataset(name_3, df_1, target_1_fake)], 0)
    ])
    def test_remove_non_binary_target_datasets(self, ds, expected_binary_n):
        da = DatasetArray(ds)
        da = deepcopy(da)
        length_1 = len(da)

        da.remove_non_binary_target_datasets()
        length_2 = len(da)

        assert length_2 <= length_1
        assert length_2 == expected_binary_n

    def test_remove_non_binary_target_datasets_nested_to_zero(self):
        da = DatasetArray([
            Dataset(name_4_nans + '_0', df_4_nans, target_1_fake),
            Dataset(name_4_nans + '_1', df_4_nans, target_4_nans),
            DatasetArray([
                Dataset(name_4_nans + '_0', df_4_nans, target_4_nans),
                Dataset(name_4_nans + '_1', df_4_nans, target_4_nans),
            ])
        ])

        assert len(da) == 3

        da = deepcopy(da)
        da.remove_non_binary_target_datasets()

        assert len(da) == 0

    def test_remove_non_binary_target_datasets_non_nested(self):
        da = DatasetArray([
            Dataset(name_4_nans + '_0', df_4_nans, target_1),
            Dataset(name_4_nans + '_1', df_4_nans, target_4_nans),
            DatasetArray([
                Dataset(name_4_nans + '_0', df_4_nans, target_1),
                Dataset(name_4_nans + '_1', df_4_nans, target_1),
            ])
        ])

        assert len(da) == 3

        da = deepcopy(da)
        da.remove_non_binary_target_datasets()

        assert len(da) == 2
        assert len(da[1]) == 2

    def test_remove_non_binary_target_datasets_not_nested_and_nested(self):
        da = DatasetArray([
            Dataset(name_4_nans + '_0', df_4_nans, target_1),
            Dataset(name_4_nans + '_1', df_4_nans, target_4_nans),
            DatasetArray([
                Dataset(name_4_nans + '_0', df_4_nans, target_1),
                Dataset(name_4_nans + '_1', df_4_nans, target_4_nans),
            ])
        ])

        assert len(da) == 3

        da = deepcopy(da)
        da.remove_non_binary_target_datasets()

        assert len(da) == 2
        assert len(da[1]) == 1

    def test_remove_non_binary_target_datasets_only_nested(self):
        da = DatasetArray([
            Dataset(name_1 + '_1', df_1, target_1),
            Dataset(name_1 + '_2', df_1, target_1),
            DatasetArray([
                Dataset(name_4_nans + '_0', df_4_nans, target_1_fake),
                Dataset(name_1 + '_1', df_1, target_1)
            ])
        ])

        assert len(da) == 3
        assert len(da[2]) == 2

        da = deepcopy(da)
        da.remove_non_binary_target_datasets()

        assert len(da) == 3
        assert len(da[2]) == 1


class TestRemoveEmpty:
    def test_remove_empty_datasets(self):
        da = DatasetArray([
            Dataset(name_1 + '_1', None, None),
            Dataset(name_1 + '_2', df_1, target_1),
            DatasetArray([
                Dataset(name_4_nans + '_0', df_4_nans, target_1_fake),
                Dataset(name_1 + '_1', df_1, target_1)
            ])
        ])

        assert len(da) == 3
        assert len(da[2]) == 2

        da = deepcopy(da)
        da.remove_non_binary_target_datasets()

        assert len(da) == 2
        assert len(da[1]) == 1

    def test_remove_empty_datasets_only_nested(self):
        da = DatasetArray([
            Dataset(name_1 + '_1', df_1, target_1),
            Dataset(name_1 + '_2', df_1, target_1),
            DatasetArray([
                Dataset(name_4_nans + '_0', None, None),
                Dataset(name_1 + '_1', df_1, target_1)
            ])
        ])

        assert len(da) == 3
        assert len(da[2]) == 2

        da = deepcopy(da)
        da.remove_non_binary_target_datasets()

        assert len(da) == 3
        assert len(da[2]) == 1

    def test_remove_empty_datasets_verbose(self, capsys):
        da = DatasetArray([
            Dataset(name_1 + '_1', df_1, target_1),
            Dataset(name_1 + '_2', df_1, target_1),
            DatasetArray([
                Dataset(name_4_nans + '_0', None, None),
                Dataset(name_1 + '_1', df_1, target_1)
            ])
        ])

        da = deepcopy(da)
        da.verbose = True
        da.remove_empty_datasets()

        captured = capsys.readouterr()
        assert captured.out.startswith(f'Empty datasets were removed from DatasetArray {da.__repr__()}\n')


@pytest.mark.parametrize('ds1,ds2', [
    (DatasetArray([Dataset(name_1, df_1, target_1)]), Dataset(name_2, df_2, target_2)),
    (DatasetArray([Dataset(name_1, df_1, target_1)]), DatasetArray([Dataset(name_2, df_3, target_3), Dataset(name_1, df_3, target_3)]))
])
def test_append(ds1, ds2):
    ds1 = deepcopy(ds1)
    ds2 = deepcopy(ds2)

    len_1 = len(ds1)
    ds1.append(ds2)
    assert len(ds1) == len_1 + 1


@pytest.mark.parametrize('ds1,ds2', [
    (DatasetArray([Dataset(name_1, df_1, target_1)]), Dataset(name_2, df_2, target_2)),
    (DatasetArray([Dataset(name_1, df_1, target_1)]), DatasetArray([Dataset(name_2, df_3, target_3), Dataset(name_1, df_3, target_3)]))
])
def test_append_list(ds1, ds2):
    ds1 = deepcopy(ds1)
    ds2 = deepcopy(ds2)
    ds3 = deepcopy(ds2)
    ds3.name += 'xx'

    len_1 = len(ds1)
    ds1.append([ds2, ds3])
    assert len(ds1) == len_1 + 2


class TestDatasetArrayFromDirectory:
    def test_create_array(self):
        try:
            DatasetArrayFromDirectory('./resources')
        except (Exception,):
            assert False

    def test_create_array_verbose(self, capsys):
        path = './resources'
        DatasetArrayFromDirectory(path, verbose=True)
        captured = capsys.readouterr()
        assert f'The files from {path} were loaded\n' in captured.out

    def test_non_existing_path(self):
        with pytest.raises(Exception):
            DatasetArrayFromDirectory('./non-existing')

    def test_not_dir(self):
        with pytest.raises(Exception):
            DatasetArrayFromDirectory('./resources/data_1.csv')



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="edgaro",
    version="1.0.1.1",
    description="Explainable imbalanceD learninG compARatOr",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/adrianstando/edgaro",
    author="Adrian Stańdo",
    author_email="adrian.j.stando@gmail.com",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: OS Independent"
    ],
    keywords="XAI, imbalance, machine learning, AI",
    packages=["edgaro"] + [f"edgaro.{item}" for item in find_packages(where="edgaro")],
    python_requires=">=3.8, <4",
    install_requires=[
        'setuptools',
        'pandas>=1.4.4',
        'numpy>=1.20.3',
        'dalex>=1.5.0',
        'imbalanced-learn>=0.10.1',
        'matplotlib>=3.4.3',
        'openml>=0.12.2',
        'pandas-profiling>=3.3.0',
        'pytest>=7.1.3',
        'scikit-learn>=1.1.0',
        'imblearn>=0.0',
        'xgboost>=1.5.0',
        'scipy>=1.7.3',
        'statsmodels>=0.13.2'
    ],
    project_urls={
        "Documentation": "https://adrianstando.github.io/edgaro",
        "Code repository": "https://github.com/adrianstando/edgaro"
    },
)



import pandas as pd
import pytest

from copy import deepcopy

from edgaro.balancing.transformer import Transformer
from edgaro.data.dataset_array import DatasetArray
from edgaro.data.dataset import Dataset
from edgaro.model.model import RandomForest, Model
from edgaro.model.model_array import ModelArray

from .resources.objects import *


@pytest.mark.parametrize('ds', [
    DatasetArray([
        Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
        Dataset(name_2, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)]))
    ]),
    DatasetArray([
        Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
        DatasetArray([
            Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
            Dataset(name_2, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)]))
        ])
    ])
])
def test_model_array(ds):
    try:
        ds.remove_nans()

        model = ModelArray(RandomForest(max_depth=1, n_estimators=1, random_state=42))
        model.fit(ds)
        model.predict(ds)
        model.predict_proba(ds)
        model.evaluate()
        str(model)
        repr(model)
    except (Exception,):
        assert False


@pytest.mark.parametrize('ds', [
    DatasetArray([
        Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
        Dataset(name_2, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)]))
    ])
])
def test_was_fitted(ds):
    ds.remove_nans()

    model = ModelArray(RandomForest(max_depth=1, n_estimators=1, random_state=42))
    assert not model.was_fitted
    model.fit(ds)
    assert model.was_fitted


@pytest.mark.parametrize('ds', [
    DatasetArray([
        Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
        Dataset(name_2, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)]))
    ])
])
def test_model_array_output(ds):
    ds.remove_nans()

    model = ModelArray(RandomForest(max_depth=1, n_estimators=1, random_state=42))
    model.fit(ds)
    y = model.predict(ds)

    assert isinstance(y, DatasetArray)
    assert len(y) == len(ds)
    for i in range(len(ds)):
        assert isinstance(y[i], Dataset)

    y = model.predict_proba(ds)
    assert isinstance(y, DatasetArray)
    assert len(y) == len(ds)
    for i in range(len(ds)):
        assert isinstance(y[i], Dataset)

    out = model.evaluate()
    assert isinstance(out, pd.DataFrame)


@pytest.mark.parametrize('ds', [
    DatasetArray([
        DatasetArray([
            Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
            Dataset(name_2, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
            Dataset(name_2 + 'xx', pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
            DatasetArray([
                Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
                Dataset(name_2, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
                Dataset(name_2 + 'xx', pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
                DatasetArray([
                    Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
                    Dataset(name_2, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
                    Dataset(name_2 + 'xx', pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
                ])
            ])
        ]),
        Dataset(name_2, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)]))
    ])
])
def test_evaluate_many_nested(ds):
    ds.remove_nans()

    model = ModelArray(RandomForest(max_depth=1, n_estimators=1, random_state=42))
    model.fit(ds)

    try:
        out = model.evaluate()
    except (Exception,):
        assert False

    assert isinstance(out, pd.DataFrame)


@pytest.mark.parametrize('ds', [
    DatasetArray([
        Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
        DatasetArray([
            Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
            Dataset(name_2, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)]))
        ])
    ])
])
def test_model_array_nested_output(ds):
    ds.remove_nans()

    model = ModelArray(RandomForest(max_depth=1, n_estimators=1, random_state=42))
    model.fit(ds)
    y = model.predict(ds)

    assert isinstance(y, DatasetArray)
    assert len(y) == len(ds)
    assert isinstance(y[0], Dataset)
    assert isinstance(y[1], DatasetArray)

    y = model.predict_proba(ds)
    assert isinstance(y, DatasetArray)
    assert len(y) == len(ds)
    assert isinstance(y[0], Dataset)
    assert isinstance(y[1], DatasetArray)

    out = model.evaluate()
    assert isinstance(out, pd.DataFrame)


@pytest.mark.parametrize('ds', [
    DatasetArray([
        Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
        Dataset(name_2, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)]))
    ])
])
def test_output_verbose(ds, capsys):
    ds.remove_nans()

    model = ModelArray(RandomForest(max_depth=1, n_estimators=1, random_state=42), verbose=True)
    model.fit(ds)
    y = model.predict(ds)
    y = model.predict_proba(ds)

    model_repr_pre = model.__repr__()
    for ele in model_repr_pre:
        if ele.isdigit():
            model_repr_pre = model_repr_pre.replace(ele, '0')

    captured = capsys.readouterr()
    assert f'ModelArray {model_repr_pre} is being fitted with {ds.name}' in captured.out
    assert f'ModelArray {model.__repr__()} was fitted with {ds.name}' in captured.out
    assert f'ModelArray {model.__repr__()} predicted on {ds.name}' in captured.out
    assert f'ModelArray {model.__repr__()} predicted probabilities on {ds.name}' in captured.out

    model.evaluate()
    captured = capsys.readouterr()
    assert f'ModelArray {model.__repr__()} is being evaluated' in captured.out
    assert f'ModelArray {model.__repr__()} was evaluated' in captured.out


@pytest.mark.parametrize('ds', [
    DatasetArray([
        Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
        Dataset(name_2, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)]))
    ])
])
@pytest.mark.parametrize('model', [
    RandomForest(max_depth=1, n_estimators=1, random_state=42)
])
def test_base_transformer_get_set(model, ds):
    array = ModelArray(model)

    assert isinstance(array.base_transformer, Model)

    try:
        ar = deepcopy(array)
        ar.base_transformer = deepcopy(model)
        ar.fit(ds)
    except (Exception,):
        assert False

    with pytest.raises(Exception):
        ar = deepcopy(array)
        tmp = deepcopy(model)
        tmp.__class__ = ModelArray
        ar.base_transformer = tmp
        x = ar.base_transformer



import numpy as np
import pandas as pd
import pytest
import re
import openml

from copy import deepcopy

from edgaro.data.dataset import Dataset, DatasetFromCSV, DatasetFromOpenML, load_mammography

from .resources.objects import *


class TestCreateObjects:
    def test_dataset(self):
        try:
            Dataset(name_1, df_1, target_1)
        except (Exception,):
            assert False

    def test_dataset_fail_df(self):
        with pytest.raises(Exception):
            Dataset(name_1, pd.concat([df_1, df_1]), target_1)

    def test_dataset_fail_target(self):
        with pytest.raises(Exception):
            Dataset(name_1, df_1, pd.concat([target_1, target_1]))

    def test_dataset_verbose(self, capsys):
        ds = Dataset(name_1, df_1, target_1, verbose=True)
        captured = capsys.readouterr()
        assert captured.out == f'Dataset {ds.__repr__()} created\n'

    def test_csv_1(self):
        try:
            ds = DatasetFromCSV(path=example_path)
        except (Exception,):
            assert False

        assert ds.target.equals(pd.Series(pd.read_csv(example_path).iloc[:, -1], name='target'))

    def test_csv_2(self):
        try:
            DatasetFromCSV(path=example_path, target=example_target)
        except (Exception,):
            assert False

    def test_csv_verbose(self, capsys):
        ds = DatasetFromCSV(path=example_path, target=example_target, verbose=True)
        captured = capsys.readouterr()
        assert captured.out.startswith(f'Data from {example_path} file is loaded\n')

    def test_openml(self):
        try:
            DatasetFromOpenML(task_id_1, apikey=APIKEY)
        except (Exception,):
            assert False

    def test_mammography(self):
        try:
            df = load_mammography()
        except (Exception,):
            assert False

        assert isinstance(df, Dataset)


@pytest.mark.parametrize('name, df, target, expected_IR, target_fake', [
    (name_1, df_1, target_1, IR_1, target_1_fake),
    (name_2, df_2, target_2, IR_2, target_2_fake)
])
class TestDatasetBasicProperties:
    def test_properties_not_none(self, name, df, target, expected_IR, target_fake):
        ds = Dataset(name, df, target)
        assert ds.data is not None
        assert ds.target is not None
        assert ds.name is not None

    def test_properties_value(self, name, df, target, expected_IR, target_fake):
        ds = Dataset(name, df, target)
        assert ds.name == name
        assert ds.data.equals(df)
        assert ds.target.equals(target)

    def test_str_repr(self, name, df, target, expected_IR, target_fake):
        try:
            ds = Dataset(name, df, target)
            str(ds)
            repr(ds)
        except (Exception,):
            assert False

    def test_imbalance_ratio(self, name, df, target, expected_IR, target_fake):
        try:
            ds = Dataset(name, df, target)
            IR = ds.imbalance_ratio
        except (Exception,):
            assert False

        assert IR == expected_IR

    def test_imbalance_ratio_three_class_target(self, name, df, target, expected_IR, target_fake):
        ds = Dataset(name, df, target_fake)
        with pytest.raises(Exception):
            IR = ds.imbalance_ratio

    def test_imbalance_ratio_empty_target(self, name, df, target, expected_IR, target_fake):
        ds = Dataset(name, df, None)
        assert ds.imbalance_ratio == 0

    def test_imbalance_ratio_one_class_target(self, name, df, target, expected_IR, target_fake):
        ds = Dataset(name, df, pd.Series(np.ones(df.shape[0])))
        assert ds.imbalance_ratio == 0

    def test_generate_report_path(self, name, df, target, expected_IR, target_fake):
        try:
            ds = Dataset(name, df, target)
            ds.generate_report(output_path=os.path.join('resources', 'out.html'), minimal=True)
        except (Exception,):
            assert False

    def test_generate_report_fail_none(self, name, df, target, expected_IR, target_fake):
        ds = Dataset(name, None, None)
        with pytest.raises(Exception):
            ds.generate_report(output_path=os.path.join('resources', 'out.html'), minimal=True)

    def test_generate_report_only_data(self, name, df, target, expected_IR, target_fake):
        ds = Dataset(name, df, None)
        try:
            ds.generate_report(output_path=os.path.join('resources', 'out.html'), minimal=True)
        except (Exception,):
            assert False

    def test_generate_report_only_target(self, name, df, target, expected_IR, target_fake):
        ds = Dataset(name, None, target)
        try:
            ds.generate_report(output_path=os.path.join('resources', 'out.html'), minimal=True)
        except (Exception,):
            assert False

    def test_check_binary(self, name, df, target, expected_IR, target_fake):
        ds = Dataset(name, df, target)
        assert ds.check_binary_classification()

    def test_check_no_binary(self, name, df, target, expected_IR, target_fake):
        tmp = Dataset(name, df, target_fake)
        assert not tmp.check_binary_classification()

    def test_equal(self, name, df, target, expected_IR, target_fake):
        ds = Dataset(name, df, target)
        tmp = Dataset(name, df, target)
        assert ds == tmp

    def test_train_test_split_before(self, name, df, target, expected_IR, target_fake):
        ds = Dataset(name, df, target)
        assert not ds.was_split

        try:
            ds.data = df
            ds.target = target
        except (Exception,):
            assert False

        assert isinstance(ds.data, pd.DataFrame)
        assert isinstance(ds.target, pd.Series)

        with pytest.raises(Exception):
            out = ds.train

        with pytest.raises(Exception):
            out = ds.test

    def test_train_test_split_after(self, name, df, target, expected_IR, target_fake):
        ds = Dataset(name, pd.concat([df for _ in range(3)]), pd.concat([target for _ in range(3)]))

        try:
            ds.train_test_split(test_size=0.3, random_state=42)
            train = ds.train
            test = ds.test
        except (Exception,):
            assert False

        assert isinstance(train, Dataset)
        assert isinstance(test, Dataset)

        assert isinstance(train.data, pd.DataFrame)
        assert isinstance(train.target, pd.Series)

        assert isinstance(test.data, pd.DataFrame)
        assert isinstance(test.target, pd.Series)

        with pytest.raises(Exception):
            ds.data = df

        with pytest.raises(Exception):
            ds.target = target

        assert ds.was_split

    def test_train_test_split_double_raise_exception(self, name, df, target, expected_IR, target_fake):
        ds = Dataset(name, pd.concat([df for _ in range(3)]), pd.concat([target for _ in range(3)]))

        try:
            ds.train_test_split(test_size=0.3, random_state=42)
            train = ds.train
            test = ds.test
        except (Exception,):
            assert False

        with pytest.raises(Exception):
            ds.train_test_split(test_size=0.3, random_state=42)

    def test_train_test_split_verbose(self, name, df, target, expected_IR, target_fake, capsys):
        ds = Dataset(name, pd.concat([df for _ in range(3)]), pd.concat([target for _ in range(3)]), verbose=True)
        ds.train_test_split(test_size=0.3, random_state=42)

        captured = capsys.readouterr()
        assert captured.out.endswith(f'Dataset {ds.__repr__()} was train-test-split\n')

    def test_custom_train_test_split(self, name, df, target, expected_IR, target_fake):
        ds = Dataset(name, None, None)
        train = Dataset(name, pd.concat([df for _ in range(3)]), pd.concat([target for _ in range(3)]))
        test = Dataset(name, pd.concat([df for _ in range(1)]), pd.concat([target for _ in range(1)]))

        ds.custom_train_test_split(
            train=train,
            test=test
        )

        assert ds.train == train
        assert ds.test == test

    def test_head(self, name, df, target, expected_IR, target_fake):
        ds = Dataset(name, df, target)
        ds_head = ds.head(2)

        assert ds_head.data.shape[0] == 2
        assert ds_head.target.shape[0] == 2


@pytest.mark.parametrize('ds1,ds2', [
    (Dataset(name_1, df_1, target_1), Dataset(name_1, df_1, target_2)),
    (Dataset(name_1, df_1, target_1), Dataset(name_1, df_2, target_1)),
    (Dataset(name_1, df_1, target_1), Dataset(name_2, df_1, target_1)),
    (Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_1)),
    (Dataset(name_1, df_1, target_1), Dataset(name_2, df_1, target_2))
])
def test_dataset_not_equal(ds1, ds2):
    assert not ds1 == ds2


@pytest.fixture(scope='module', params=[example_path])
def dataset_csv(request):
    return DatasetFromCSV(request.param)


class TestCSV:
    def test_calculate_imbalance_ratio(self, dataset_csv):
        try:
            IR = dataset_csv.imbalance_ratio
        except (Exception,):
            assert False

    def test_check_binary(self, dataset_csv):
        assert dataset_csv.check_binary_classification()


@pytest.fixture(scope='module', params=[task_id_1, task_id_2])
def dataset_openml(request):
    return DatasetFromOpenML(request.param, apikey=APIKEY)


class TestOpenML:
    def test_calculate_imbalance_ratio(self, dataset_openml):
        try:
            IR = dataset_openml.imbalance_ratio
        except (Exception,):
            assert False

    def test_check_binary(self, dataset_openml):
        assert dataset_openml.check_binary_classification()

    def test_description_openml(self, dataset_openml):
        out = dataset_openml.openml_description()
        pattern = '^Name: .* Description: .*$'
        assert re.match(pattern, out.replace('\n', ' '))


@pytest.mark.parametrize('task', [
    task_id_1
])
def test_dataset_openml_verbose(task, capsys):
    ds = DatasetFromOpenML(task, apikey=APIKEY, verbose=True)
    captured = capsys.readouterr()
    assert captured.out.startswith(f'Dataset from OpenML with id {str(task)} was downloaded\n')


@pytest.mark.parametrize('task', [
    task_id_1
])
def test_apikey_fail_none(task):
    tmp_code = openml.config.apikey
    openml.config.apikey = ''

    with pytest.raises(Exception):
        DatasetFromOpenML(task, apikey=None)

    openml.config.apikey = tmp_code


@pytest.mark.parametrize('task', [
    task_id_1
])
def test_apikey_with_apikey_empty_openml(task):
    tmp_code = openml.config.apikey
    openml.config.apikey = ''

    try:
        DatasetFromOpenML(task, apikey=APIKEY)
    except (Exception,):
        assert False
    finally:
        openml.config.apikey = tmp_code


@pytest.mark.parametrize('name, df, target, col_thresh_nan_remove, data_shape_nan, target_shape_nan', [
    (name_4_nans, deepcopy(df_4_nans), deepcopy(target_4_nans), 0.9, (1, 2), (1,)),
    (name_4_nans, deepcopy(df_4_nans), deepcopy(target_2), 1, (2, 2), (2,))
])
def test_output_shape(name, df, target, col_thresh_nan_remove, data_shape_nan, target_shape_nan):
    # noinspection PyTypeChecker
    ds = Dataset(name, df, target)
    ds.remove_nans(col_thresh=col_thresh_nan_remove)

    assert ds.data.shape == data_shape_nan
    assert ds.target.shape == target_shape_nan


@pytest.mark.parametrize('name, df, target, n_rows', [
    (name_1, deepcopy(df_1), deepcopy(target_1), 3),
    (name_2, deepcopy(df_2), deepcopy(target_2), 3)
])
def test_remove_outliers(name, df, target, n_rows):
    ds = Dataset(name, df, target)
    ds.data = pd.concat([ds.data, pd.DataFrame({'a': [0, 0, 0, 1000], 'b': ['a', 'b', 'p', 'q']})])
    ds.target = pd.concat([ds.target, pd.Series([0, 0, 1, 1], name='target')])

    assert ds.data.shape[0] == n_rows + 4
    assert ds.target.shape[0] == n_rows + 4

    ds.remove_outliers(n_std=1)

    assert ds.data.shape[0] == n_rows + 3
    assert ds.target.shape[0] == n_rows + 3



import warnings
import pandas as pd
import numpy as np

from abc import ABC, abstractmethod
from typing import Dict, Protocol, Any, Tuple, Optional, List, Union
from copy import deepcopy
from imblearn.under_sampling import RandomUnderSampler as RUS
from imblearn.over_sampling import RandomOverSampler as ROS
from imblearn.over_sampling import SMOTE as SM_C, SMOTENC as SM_NC, SMOTEN as SM_N
from sklearn.utils.validation import check_is_fitted
from sklearn.exceptions import NotFittedError

from edgaro.data.dataset import Dataset
from edgaro.data.dataset_array import DatasetArray
from edgaro.base.base_transformer import BaseTransformer
from edgaro.base.utils import print_unbuffered


class Transformer(BaseTransformer, ABC):
    """
    The abstract class to define balancing transformations for a single Dataset.

    Parameters
    ----------
    name_sufix : str, default='_transformed'
        Sufix to be set to a transformed Dataset.
    verbose : bool, default=False
        Print messages during calculations.

    Attributes
    ----------
    name_sufix : str
        Sufix to be set to a transformed Dataset.
    verbose : bool
        Print messages during calculations.
    """

    def __init__(self, name_sufix: str = '_transformed', verbose: bool = False) -> None:
        super().__init__()
        self.name_sufix = name_sufix
        self.__was_fitted = False
        self.verbose = verbose

    def fit(self, dataset: Dataset) -> None:
        """
        Fit the transformer.

        Parameters
        ----------
        dataset : Dataset
            The object to fit Transformer on.
        """
        if dataset.target is None:
            raise Exception('Target data is not provided!')
        if dataset.data is None:
            raise Exception('Data in dataset is not provided!')

        d = dataset.train if dataset.was_split else dataset
        self._fit(d)
        self.__was_fitted = True

        if self.verbose:
            print_unbuffered(f'Transformer {self.__repr__()} was fitted with {dataset.name}')

    @abstractmethod
    def _fit(self, dataset: Dataset) -> None:
        pass

    @property
    def was_fitted(self) -> bool:
        """
        The information whether the Transformer was fitted.

        Returns
        -------
        bool
        """
        return self.__was_fitted

    def transform(self, dataset: Dataset) -> Union[Dataset, DatasetArray]:
        """
        Transform the object.

        Parameters
        ----------
        dataset : Dataset
            The object to be transformed.

        Returns
        -------
        Dataset, DatasetArray
            The transformed object.
        """
        ds = deepcopy(dataset)
        if ds.was_split:
            d = ds.train
            new_train = self._transform(d)
            out = Dataset(name=ds.name + self.name_sufix, dataframe=None, target=None)
            out.custom_train_test_split(
                train=new_train,
                test=ds.test
            )
        else:
            out = self._transform(ds)

        if self.verbose:
            print_unbuffered(f'Transformer {self.__repr__()} transformed with {dataset.name}')

        if dataset.majority_class_label is None:
            names, counts = np.unique(dataset.target, return_counts=True)
            out.majority_class_label = names[1 if counts[0] < counts[1] else 0]
        else:
            out.majority_class_label = dataset.majority_class_label

        return out

    @abstractmethod
    def _transform(self, dataset: Dataset) -> Dataset:
        pass

    @abstractmethod
    def set_params(self, **params) -> None:
        """
        Set params for Transformer.

        Parameters
        ----------
        params : dict
            The parameters to be set.
        """
        pass

    @abstractmethod
    def get_params(self) -> Union[Dict, List]:
        """
        Get parameters of Transformer.

        Returns
        -------
        Dict, list
            The parameters.
        """
        pass

    def set_dataset_suffixes(self, name_sufix: Union[str, list]) -> None:
        """
        Set sufix to be set to transformed Dataset.

        Parameters
        ----------
        name_sufix : str, list
            Sufix to be set to a transformed Dataset.
        """
        self.name_sufix = name_sufix

    def __str__(self) -> str:
        return f"Transformer {self.__class__.__name__}"

    def __repr__(self) -> str:
        return f"<Transformer {self.__class__.__name__}>"


class ImblearnProtocol(Protocol):
    """
    A Protocol to define the expected structure of a Transformer from `imblearn` library.
    """
    def fit(self, X, y) -> Any:
        ...

    def fit_resample(self, X, y) -> Tuple[pd.DataFrame, pd.Series]:
        ...

    def get_params(self) -> Dict:
        ...

    def set_params(self, **params) -> Any:
        ...


class TransformerFromIMBLEARN(Transformer):
    """ Create balancing Transformer from Transformer implemented in `imblearn` library.

    Parameters
    ----------
    transformer : ImblearnProtocol
        Transformer from imblearn` library.
    name_sufix : str, default='_transformed'
        Sufix to be set to a transformed Dataset.
    verbose : bool, default=False
        Print messages during calculations.

    Examples
    ----------

    >>> from test.resources.objects import *
    >>> from imblearn.under_sampling import RandomUnderSampler
    >>> dataset = Dataset(name_1, df_1, target_1)
    >>> transformator = TransformerFromIMBLEARN(RandomUnderSampler(sampling_strategy=1, random_state=42))
    >>> transformator.fit(dataset)
    >>> transformator.transform(dataset)
    """

    def __init__(self, transformer: ImblearnProtocol, name_sufix: str = '_transformed', verbose: bool = False) -> None:
        self.__transformer = transformer
        super().__init__(name_sufix=name_sufix, verbose=verbose)

    def _change_transformer(self, transformer: ImblearnProtocol) -> None:
        try:
            check_is_fitted(transformer)
            warnings.warn('Transformer was not changed! The existing one was already fitted.')
        except NotFittedError:
            self.__transformer = transformer

    def _fit(self, dataset: Dataset) -> None:
        self.__transformer.fit(dataset.data, dataset.target)

    def transform(self, dataset: Dataset) -> Dataset:
        """
        Transform the object.

        Parameters
        ----------
        dataset : Dataset
            The object to be transformed.

        Returns
        -------
        Dataset
            The transformed object.
        """
        return super().transform(dataset)

    def _transform(self, dataset: Dataset) -> Dataset:
        X, y = self.__transformer.fit_resample(dataset.data, dataset.target)
        name = dataset.name + self.name_sufix
        return Dataset(name=name, dataframe=X, target=y)

    def get_imblearn_transformer(self) -> ImblearnProtocol:
        """
        Get the base transformer object from `imblearn` library.

        Returns
        -------
        ImblearnProtocol
        """
        return self.__transformer

    def set_params(self, **params) -> None:
        """
        Set params for Transformer.

        The function allows using `imbalance_ratio` and `IR` parameters, which are transformed to `sampling_strategy`
        parameter in `imblearn` Transformer.

        Parameters
        ----------
        params : dict
            The parameters to be set.
        """
        if 'imbalance_ratio' in params.keys():
            x = params.pop('imbalance_ratio')
            params['sampling_strategy'] = 1/x
        elif 'IR' in params.keys():
            IR = params.pop('IR')
            params['sampling_strategy'] = 1 / IR
        self.__transformer.set_params(**params)

    def get_params(self) -> Dict:
        """
        Get parameters of Transformer.

        Returns
        -------
        Dict, list
            The parameters.
        """
        return self.__transformer.get_params()

    def set_dataset_suffixes(self, name_sufix: str) -> None:
        """
        Set sufix to be set to transformed Dataset.

        Parameters
        ----------
        name_sufix : str
            Sufix to be set to a transformed Dataset.
        """
        super().set_dataset_suffixes(name_sufix)


class RandomUnderSampler(TransformerFromIMBLEARN):
    """ Create Random Under Sampling transformer.

    Parameters
    ----------
    imbalance_ratio : float, default=1
        Imbalance Ratio after transformations.
    name_sufix : str, default='_transformed'
        Sufix to be set to a transformed Dataset.
    verbose : bool, default=False
        Print messages during calculations.
    random_state : int, optional
        Random state seed.
    *args : tuple, optional
        Additional parameter for Random Under Sampling transformer from `imblearn`.
    **kwargs : dict, optional
        Additional parameter for Random Under Sampling transformer from `imblearn`.
    """

    def __init__(self, imbalance_ratio: float = 1, name_sufix: str = '_transformed',
                 random_state: int = None, verbose: bool = False, *args, **kwargs) -> None:
        transformer = RUS(sampling_strategy=1 / imbalance_ratio, random_state=random_state, *args, **kwargs)
        super().__init__(transformer=transformer, name_sufix=name_sufix, verbose=verbose)

    def set_params(self, **params) -> None:
        return super().set_params(**params)


class RandomOverSampler(TransformerFromIMBLEARN):
    """ Create Random Over Sampling transformer.

    Parameters
    ----------
    imbalance_ratio : float, default=1
        Imbalance Ratio after transformations.
    name_sufix : str, default='_transformed'
        Sufix to be set to a transformed Dataset.
    verbose : bool, default=False
        Print messages during calculations.
    random_state : int, optional
        Random state seed.
    *args : tuple, optional
        Additional parameter for Random Over Sampling transformer from `imblearn`.
    **kwargs : dict, optional
        Additional parameter for Random Over Sampling transformer from `imblearn`.
    """

    def __init__(self, imbalance_ratio: float = 1, name_sufix: str = '_transformed',
                 random_state: int = None, verbose: bool = False, *args, **kwargs) -> None:
        transformer = ROS(sampling_strategy=1 / imbalance_ratio, random_state=random_state, *args, **kwargs)
        super().__init__(transformer=transformer, name_sufix=name_sufix, verbose=verbose)

    def set_params(self, **params) -> None:
        return super().set_params(**params)


class SMOTE(TransformerFromIMBLEARN):
    """ Create SMOTE/SMOTENC/SMOTEN transformer.

    The method works also with categorical variables. The method guesses that columns of types 'category', 'object',
    'int' are always categorical. Keep that in mind before using this Transformer!

    Parameters
    ----------
    imbalance_ratio : float, default=1
        Imbalance Ratio after transformations.
    name_sufix : str, default='_transformed'
        Sufix to be set to a transformed Dataset.
    verbose : bool, default=False
        Print messages during calculations.
    random_state : int, optional
        Random state seed.
    *args : tuple, optional
        Additional parameter for SMOTE/SMOTENC/SMOTEN transformer from `imblearn`.
    **kwargs : dict, optional
        Additional parameter for SMOTE/SMOTENC/SMOTEN transformer from `imblearn`.
    """

    def __init__(self, imbalance_ratio: float = 1, name_sufix: str = '_transformed', random_state: int = None,
                 columns_categorical: Optional[List[str]] = None, verbose: bool = False, *args, **kwargs) -> None:
        self.__sampling_strategy = 1 / imbalance_ratio
        self.__random_state = random_state
        self.__args = args
        self.__kwargs = kwargs
        self.__given_columns_categorical = columns_categorical
        transformer = SM_C(sampling_strategy=1 / imbalance_ratio, random_state=random_state, *args, **kwargs)
        super().__init__(transformer=transformer, name_sufix=name_sufix, verbose=verbose)

    def _fit(self, dataset: Dataset) -> None:
        if self.__given_columns_categorical is not None:
            columns_categorical = self.__given_columns_categorical
        else:
            columns_categorical = list(dataset.data.select_dtypes(include=['category', 'object', 'int']))

        if len(columns_categorical) > 0:
            columns = dataset.data.columns
            if len(columns_categorical) == len(columns):
                super()._change_transformer(
                    SM_N(sampling_strategy=1 / self.__sampling_strategy, random_state=self.__random_state,
                         *self.__args, **self.__kwargs))
            else:
                all_columns = list(dataset.data.columns)
                categorical_indexes = [all_columns.index(c) for c in columns_categorical]
                super()._change_transformer(
                    SM_NC(categorical_indexes, sampling_strategy=1 / self.__sampling_strategy,
                          random_state=self.__random_state, *self.__args, **self.__kwargs))

        super()._fit(dataset)

    def set_params(self, **params) -> None:
        return super().set_params(**params)



from __future__ import annotations

import numpy as np

from typing import List, Dict, Optional, Union
from copy import deepcopy
from imblearn.over_sampling import ADASYN, BorderlineSMOTE, KMeansSMOTE, SVMSMOTE
from imblearn.under_sampling import ClusterCentroids, NearMiss
from imblearn.combine import SMOTEENN, SMOTETomek

from edgaro.balancing.transformer import TransformerFromIMBLEARN
from edgaro.balancing.transformer_array import TransformerArray
from edgaro.data.dataset import Dataset
from edgaro.data.dataset_array import DatasetArray
from edgaro.balancing.transformer import Transformer, RandomUnderSampler, RandomOverSampler, SMOTE
from edgaro.base.utils import print_unbuffered


class NestedAutomaticTransformer(Transformer):
    """
    Create NestedAutomaticTransformer.

    This object creates an object that behaves like a Transformer, but transforms a single Dataset using not only one,
    but more methods. Moreover, you do not specify Imbalance Ratio values, but they are automatically calculated. You
    pass only an argument `n_per_method` and the class will balance the Dataset
    with that number of Imbalance Raio values.

    Note: If you use NestedAutomaticTransformer (or children class) as a parameter to TransformerArray, it is advisable
    to pass parameter `set_suffixes=False` in TransformerArray object. Otherwise, the suffixes will be distorted.

    Parameters
    ----------
    base_transformers : list(Transformer)
        List of base transformers.
    base_transformers_names : list(str)
        List of names of base_transformers.
    keep_original_dataset : bool, default=False
        Keep the original Dataset after transformations or not.
    result_array_sufix str, default='_automatic_transformed_array'
        Suffix of the transformed DatasetArray.
    n_per_method : int, default=5
        Number of intermediate Imbalance Ratio values.
    random_state : int, optional, default=None,
        Random state seed.
    IR_round_precision : int, default=2
        Round precision of Imbalance Ratio when printing.
    min_samples_to_modify : int, optional, default=None
        Minimal number of samples to modify to create an intermediate Imbalance Ratio. If the number of modified
        observations is less than this number, the `n_per_method` parameter will be modified.
    verbose : bool, default=False
        Print messages during calculations.

    Attributes
    ----------
    keep_original_dataset : bool, default=False
        Keep the original Dataset after transformations or not.
    n_per_method : int, default=5
        Number of intermediate Imbalance Ratio values.
    random_state : int, optional, default=None,
        Random state seed.
    result_array_sufix str, default='_automatic_transformed_array'
        Suffix of the transformed DatasetArray.
    IR_round_precision : int, default=2
        Round precision of Imbalance Ratio when printing.
    min_samples_to_modify : int, optional, default=None
        Minimal number of samples to modify to create an intermediate Imbalance Ratio. If the number of modified
        observations is less than this number, the `n_per_method` parameter will be modified.
    verbose : bool, default=False
        Print messages during calculations.

    """
    def __init__(self, base_transformers: List[Transformer], base_transformers_names: List[str],
                 keep_original_dataset: bool = False, result_array_sufix: str = '_automatic_transformed_array',
                 n_per_method: int = 5, random_state: Optional[int] = None, IR_round_precision: int = 2,
                 min_samples_to_modify: Optional[int] = None, verbose: bool = False) -> None:
        self.keep_original_dataset = keep_original_dataset
        self.n_per_method = n_per_method
        self.random_state = random_state
        self.result_array_sufix = result_array_sufix
        self.IR_round_precision = IR_round_precision
        self.__transformers = None
        self.__was_fitted = False
        self.min_samples_to_modify = min_samples_to_modify
        self.verbose = verbose
        self.__base_transformers = base_transformers
        self.__base_transformers_names = base_transformers_names

        super(Transformer, self).__init__()

    def __find_IR_values(self, dataset: Dataset) -> List[float]:
        IR = dataset.imbalance_ratio
        n_per_method = self.n_per_method
        IR_range = IR - 1
        IR_step = IR_range / n_per_method

        if self.min_samples_to_modify is not None:
            _, counts = np.unique(dataset.target, return_counts=True)
            n_rows_minority = min(counts)
            n_rows_majority = max(counts)

            n_rows_modified_under = IR_step * n_rows_minority
            n_rows_modified_over = (n_rows_majority * n_rows_minority) / (
                        n_rows_majority - IR_step * n_rows_minority) - n_rows_minority
            min_n_rows_modified = min(n_rows_modified_under, n_rows_modified_over)

            if min_n_rows_modified < self.min_samples_to_modify:
                IR_step_under = self.min_samples_to_modify / n_rows_minority
                IR_step_over = n_rows_majority * (
                            1 / n_rows_minority - 1 / (n_rows_minority + self.min_samples_to_modify))

                IR_step = max(IR_step_under, IR_step_over)
                n_per_method = int((IR - 1) // IR_step + 1)

                if round(IR_step, 3) == round(IR_range, 3):
                    n_per_method = 1

        IR_values = [
            IR - i * IR_step if IR - i * IR_step >= 1
            else 1
            for i in range(1, int(n_per_method) + 1)
        ]

        return IR_values

    def fit(self, dataset: Dataset) -> None:
        """
        Fit the transformer.

        Parameters
        ----------
        dataset : Dataset
            The object to fit Transformer on.
        """
        IR_values = self.__find_IR_values(dataset)

        for i in range(len(self.__base_transformers)):
            try:
                self.__base_transformers[i] = deepcopy(self.__base_transformers[i])
                self.__base_transformers[i].set_params(random_state=self.random_state)
            except (Exception,):
                pass

        self.__transformers = [
            TransformerArray(
                self.__base_transformers[i],
                parameters=[
                    {'IR': ir}
                    for ir in IR_values
                ],
                dataset_suffixes=[
                    f'_transformed_{self.__base_transformers_names[i]}_IR='
                    f'{round(ir, self.IR_round_precision)}='
                    f'{round(ir/dataset.imbalance_ratio, self.IR_round_precision)}BASE_IR='
                    f'{round((ir-1)/(dataset.imbalance_ratio-1), self.IR_round_precision)}IR_RANGE'
                    for ir in IR_values
                ],
                result_array_sufix=f'_{self.__base_transformers_names[i]}'
            )
            for i in range(len(self.__base_transformers))
        ]

        for transformer in self.__transformers:
            transformer.fit(dataset)

        self.__was_fitted = True

        if self.verbose:
            print_unbuffered(f'AutomaticTransformerArray {self.__repr__()} was fitted with {dataset.name}')

    def transform(self, dataset: Dataset) -> DatasetArray:
        """
        Transform the object.

        Parameters
        ----------
        dataset : Dataset
            The object to be transformed.

        Returns
        -------
        DatasetArray
            The transformed object.
        """
        if self.__transformers is None:
            return DatasetArray([])
        else:
            out = DatasetArray(
                [t.transform(dataset) for t in self.__transformers],
                name=dataset.name + self.result_array_sufix
            )
            if self.keep_original_dataset:
                out.append(dataset)

            if self.verbose:
                print_unbuffered(f'AutomaticTransformerArray {self.__repr__()} transformed {dataset.name}')

            return out

    def set_dataset_suffixes(self, name_sufix: Union[str, List[Union[str, List[str]]]]) -> None:
        """
        Set suffixes to be set to transformed Dataset.

        Parameters
        ----------
        name_sufix : str, list
            Suffixes to be set to a transformed Dataset.
        """
        if isinstance(self.__transformers, list) and np.all(
                [isinstance(t, TransformerArray) for t in self.__transformers]):
            for t in self.__transformers:
                t.set_dataset_suffixes(name_sufix)

    def get_dataset_suffixes(self) -> Optional[Union[str, List[str]]]:
        """
        Get suffixes for transformed Dataset.

        Returns
        -------
        str, list
            Suffixes for a transformed Dataset.
        """
        if isinstance(self.__transformers, list) and np.all(
                [isinstance(t, TransformerArray) for t in self.__transformers]):
            return [t.get_dataset_suffixes() for t in self.__transformers]
        else:
            return []

    def set_params(self, **params) -> None:
        """
        Set params for Transformer.

        Parameters
        ----------
        params : dict
            The parameters to be set.
        """
        if isinstance(self.__transformers, list) and np.all(
                [isinstance(t, TransformerArray) for t in self.__transformers]):
            for i in range(len(self.__transformers)):
                tmp = {}
                for key, val in params.items():
                    tmp[key] = val[i]
                self.__transformers[i].set_params(**tmp)

    def get_params(self) -> List[Union[Dict, List]]:
        """
        Get parameters of Transformer.

        Returns
        -------
        Dict, list
            The parameters.
        """
        if isinstance(self.__transformers, list) and np.all(
                [isinstance(t, TransformerArray) for t in self.__transformers]):
            return [t.parameters for t in self.__transformers]
        else:
            return []

    @property
    def was_fitted(self) -> bool:
        return self.__was_fitted

    @property
    def transformers(self) -> List[Union[Transformer, TransformerArray, List]]:
        """
        All the Transformer objects used by this object.

        Returns
        -------
        list[Transformer, TransformerArray, list]
        """
        if self.__transformers is None:
            return []
        else:
            return self.__transformers

    @transformers.setter
    def transformers(self, val: List[Union[Transformer, TransformerArray, List]]) -> None:
        if not self.was_fitted:
            self.__transformers = val
        else:
            raise Exception('Transformers were not set since Transformer has already been fitted!')

    @property
    def base_transformer(self) -> Union[Transformer, TransformerArray, List]:
        """
        Base transformers for creation of this object.

        Returns
        -------
        list[Transformer, TransformerArray, list]
        """
        if self.__base_transformers is None:
            return []
        else:
            return self.__base_transformers

    @base_transformer.setter
    def base_transformer(self, val: Union[Transformer, TransformerArray, List]) -> None:
        if not self.was_fitted:
            self.__base_transformers = val
        else:
            raise Exception('BaseTransformers were not set since Transformer has already been fitted!')

    def _fit(self, dataset: Dataset) -> None:
        pass

    def _transform(self, dataset: Dataset) -> Dataset:
        return Dataset('', None, None)

    def __len__(self) -> int:
        if self.__transformers is None:
            return 0
        else:
            return len(self.__transformers)

    def __getitem__(self, key: Union[int, List[int]]) -> Optional[
                    Union[Optional[Transformer], TransformerArray, List[Union[Optional[Transformer], TransformerArray]]]]:
        if self.__transformers is None:
            return None
        if isinstance(key, list):
            out = [self.__getitem__(k) for k in key]
            if len(out) == 0:
                return None
            else:
                return out
        elif isinstance(key, int):
            if key <= len(self.__transformers):
                return self.__transformers[key]
        return None

    def __str__(self) -> str:
        out = f"NestedTransformer transformers" + "\n"
        if self.__transformers is not None:
            for t in self.__transformers:
                out += t.__str__()
                out += '\n'
        return out

    def __repr__(self) -> str:
        return "<NestedTransformer transformer>"


class BasicAutomaticTransformer(NestedAutomaticTransformer):
    """
    Create BasicAutomaticTransformer.

    This object contains three most popular methods - Random Under Sampler, Random Over Sampler and SMOTE.

    This class can be used for both continuous (numerical) and categorical data.

    Parameters
    ----------
    keep_original_dataset : bool, default=False
        Keep the original Dataset after transformations or not.
    result_array_sufix str, default='_automatic_transformed_array'
        Suffix of the transformed DatasetArray.
    n_per_method : int, default=5
        Number of intermediate Imbalance Ratio values.
    random_state : int, optional, default=None,
        Random state seed.
    IR_round_precision : int, default=2
        Round precision of Imbalance Ratio when printing.
    min_samples_to_modify : int, optional, default=None
        Minimal number of samples to modify to create an intermediate Imbalance Ratio. If the number of modified
        observations is less than this number, the `n_per_method` parameter will be modified.
    verbose : bool, default=False
        Print messages during calculations.

    """
    def __init__(self, keep_original_dataset: bool = False, result_array_sufix: str = '_automatic_transformed_array',
                 n_per_method: int = 5, random_state: Optional[int] = None, IR_round_precision: int = 2,
                 min_samples_to_modify: Optional[int] = None, verbose: bool = False) -> None:

        super().__init__(
            base_transformers=[RandomUnderSampler(), RandomOverSampler(), SMOTE()],
            base_transformers_names=['UNDERSAMPLING__Random', 'OVERSAMPLING__Random', 'OVERSAMPLING__SMOTE'],
            keep_original_dataset=keep_original_dataset, result_array_sufix=result_array_sufix,
            n_per_method=n_per_method, random_state=random_state, IR_round_precision=IR_round_precision,
            min_samples_to_modify=min_samples_to_modify, verbose=verbose)

    def __str__(self) -> str:
        out = f"BasicAutomaticTransformer transformers" + "\n"
        if self.transformers is not None:
            for t in self.transformers:
                out += t.__str__()
                out += '\n'
        return out

    def __repr__(self) -> str:
        return "<BasicAutomaticTransformer transformer>"


class ExtensionAutomaticTransformer(NestedAutomaticTransformer):
    """
    Create ExtensionAutomaticTransformer.

    This object contains three more complex methods implemented in imblearn. There are methods used for oversampling
    (BorderlineSMOTE), undersampling (NearMiss) and there is also a hybrid method (SMOTETomek).

    This class can be used only for continuous (numerical) data.

    Parameters
    ----------
    keep_original_dataset : bool, default=False
        Keep the original Dataset after transformations or not.
    result_array_sufix str, default='_automatic_transformed_array'
        Suffix of the transformed DatasetArray.
    n_per_method : int, default=5
        Number of intermediate Imbalance Ratio values.
    random_state : int, optional, default=None,
        Random state seed.
    IR_round_precision : int, default=2
        Round precision of Imbalance Ratio when printing.
    min_samples_to_modify : int, optional, default=None
        Minimal number of samples to modify to create an intermediate Imbalance Ratio. If the number of modified
        observations is less than this number, the `n_per_method` parameter will be modified.
    verbose : bool, default=False
        Print messages during calculations.

    """
    def __init__(self, keep_original_dataset: bool = False, result_array_sufix: str = '_automatic_transformed_array',
                 n_per_method: int = 5, random_state: Optional[int] = None, IR_round_precision: int = 2,
                 min_samples_to_modify: Optional[int] = None, verbose: bool = False) -> None:

        super().__init__(
            base_transformers=[
                TransformerFromIMBLEARN(NearMiss()),
                TransformerFromIMBLEARN(BorderlineSMOTE()),
                TransformerFromIMBLEARN(SMOTETomek())
            ],
            base_transformers_names=[
                'UNDERSAMPLING__Near_Miss',
                'OVERSAMPLING__Borderline_SMOTE',
                'COMBINED__SMOTE_Tomek'
            ],
            keep_original_dataset=keep_original_dataset, result_array_sufix=result_array_sufix,
            n_per_method=n_per_method, random_state=random_state, IR_round_precision=IR_round_precision,
            min_samples_to_modify=min_samples_to_modify, verbose=verbose)

    def __str__(self) -> str:
        out = f"ExtensionAutomaticTransformer transformers" + "\n"
        if self.transformers is not None:
            for t in self.transformers:
                out += t.__str__()
                out += '\n'
        return out

    def __repr__(self) -> str:
        return "<ExtensionAutomaticTransformer transformer>"


class AutomaticTransformer(NestedAutomaticTransformer):
    """
    Create AutomaticTransformer.

    This object contains six methods implemented in imblearn. There are methods used for oversampling
    (RandomOverSampling, BorderlineSMOTE), undersampling (RandomUnderSampling, NearMiss) and there is also a
    hybrid method (SMOTETomek).

    This class can be used only for continuous (numerical) data.

    Parameters
    ----------
    keep_original_dataset : bool, default=False
        Keep the original Dataset after transformations or not.
    result_array_sufix str, default='_automatic_transformed_array'
        Suffix of the transformed DatasetArray.
    n_per_method : int, default=5
        Number of intermediate Imbalance Ratio values.
    random_state : int, optional, default=None,
        Random state seed.
    IR_round_precision : int, default=2
        Round precision of Imbalance Ratio when printing.
    min_samples_to_modify : int, optional, default=None
        Minimal number of samples to modify to create an intermediate Imbalance Ratio. If the number of modified
        observations is less than this number, the `n_per_method` parameter will be modified.
    verbose : bool, default=False
        Print messages during calculations.

    """
    def __init__(self, keep_original_dataset: bool = False, result_array_sufix: str = '_automatic_transformed_array',
                 n_per_method: int = 5, random_state: Optional[int] = None, IR_round_precision: int = 2,
                 min_samples_to_modify: Optional[int] = None, verbose: bool = False) -> None:

        super().__init__(
            base_transformers=[
                RandomUnderSampler(),
                TransformerFromIMBLEARN(NearMiss()),
                RandomOverSampler(),
                SMOTE(),
                TransformerFromIMBLEARN(BorderlineSMOTE()),
                TransformerFromIMBLEARN(SMOTETomek())
            ],
            base_transformers_names=[
                'UNDERSAMPLING__Random',
                'UNDERSAMPLING__Near_Miss',
                'OVERSAMPLING__Random',
                'OVERSAMPLING__SMOTE',
                'OVERSAMPLING__Borderline_SMOTE',
                'COMBINED__SMOTE_Tomek'
            ],
            keep_original_dataset=keep_original_dataset, result_array_sufix=result_array_sufix,
            n_per_method=n_per_method, random_state=random_state, IR_round_precision=IR_round_precision,
            min_samples_to_modify=min_samples_to_modify, verbose=verbose)

    def __str__(self) -> str:
        out = f"AutomaticTransformer transformers" + "\n"
        if self.transformers is not None:
            for t in self.transformers:
                out += t.__str__()
                out += '\n'
        return out

    def __repr__(self) -> str:
        return "<AutomaticTransformer transformer>"
